# Load packages
import matplotlib.pyplot as plt
import pandas as pd
from scipy.stats import chi2_contingency
import statsmodels.api as sm
import seaborn as sns
import textwrap

# Declare function
def ChiSquareTestOfIndependenceFromTable(contingency_table,
                                         column_wise_variable_name,
                                         show_contingency_tables=True,
                                         show_plot=True,
                                         # Plot formatting arguments
                                         color_palette="Paired",
                                         fill_transparency=0.8,
                                         figure_size=(6, 8),
                                         # Text formatting arguments
                                         title_for_plot="Chi-Square Test of Independence",
                                         subtitle_for_plot="Shows observed vs. expected counts",
                                         caption_for_plot="Expected counts are based on the null hypothesis of no association between the two variables.",
                                         data_source_for_plot=None,
                                         x_indent=-0.95,
                                         title_y_indent=1.15,
                                         title_font_size=14,
                                         subtitle_y_indent=1.1,
                                         subtitle_font_size=11,
                                         caption_y_indent=-0.15,
                                         caption_font_size=8,
                                         decimal_places_for_data_label=1):
    """
    Performs a chi-square test of independence on a contingency table and returns the observed and expected counts.

    Args:
        contingency_table (pandas.DataFrame): A contingency table with the row-wise variable as the index and the column-wise variable as the columns.
        column_wise_variable_name (str): The name of the column-wise variable.
        show_contingency_tables (bool, optional): Whether to show the observed and expected contingency tables. Defaults to True.
        show_plot (bool, optional): Whether to show a clustered bar chart of the observed and expected values. Defaults to True.
        color_palette (str, optional): The color palette to use for the plot. Defaults to "Set1".
        fill_transparency (float, optional): The transparency of the bars in the plot. Defaults to 0.8.
        figure_size (tuple, optional): The size of the plot. Defaults to (6, 6).
        title_for_plot (str, optional): The title of the plot. Defaults to "Chi-Square Test of Independence".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to "Shows observed vs. expected counts".
        caption_for_plot (str, optional): The caption of the plot. Defaults to "Expected counts are based on the null hypothesis of no association between the two variables.".
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        x_indent (float, optional): The x-axis indent of the plot. Defaults to -0.95.
        title_y_indent (float, optional): The y-axis indent of the title of the plot. Defaults to 1.15.
        subtitle_y_indent (float, optional): The y-axis indent of the subtitle of the plot. Defaults to 1.1.
        caption_y_indent (float, optional): The y-axis indent of the caption of the plot. Defaults to -0.15.
        decimal_places_for_data_label (int, optional): The number of decimal places to show in the data labels of the plot. Defaults to 1.

    Returns:
        pandas.DataFrame: A DataFrame containing the observed and expected counts, as well as the difference between the observed and expected counts.
    """
    
    # Get row-wise variable names
    row_wise_variable_name = contingency_table.index.name
    
    # Create contingency table in Python
    contingency_table = sm.stats.Table(contingency_table)
    
    # Perform chi-square test of independence
    results = chi2_contingency(contingency_table.table_orig)
    
    # Print results of test
    print("Chi-square test statistic: " + str(round(results[0], 3)))
    print("p-value: " + str(round(results[1], 3)))
    print("Degrees of freedom: " + str(round(results[2], 3)))
    
    # Show contingency table with margins
    if show_contingency_tables:
        print("")
        print("Observed:")
        print(contingency_table.table_orig)
        print("")
        print("Expected:")
        print(contingency_table.fittedvalues.round(2))
        
    # Combine observed and expected counts
    df_counts = pd.DataFrame(contingency_table.table_orig)
    df_counts = df_counts.reset_index()
    df_counts = df_counts.melt(
        id_vars=row_wise_variable_name,
        var_name=column_wise_variable_name,
        value_name='Observed'
    )
    df_temp = pd.DataFrame(contingency_table.fittedvalues.round(2))
    df_temp = df_temp.reset_index()
    df_temp = df_temp.melt(
        id_vars=row_wise_variable_name,
        var_name=column_wise_variable_name,
        value_name='Expected'
    )
    df_counts = df_counts.merge(
        df_temp,
        how='left',
        on=[row_wise_variable_name, column_wise_variable_name]
    )
    
    # Convert to long-form
    df_counts = df_counts.melt(
        id_vars=[row_wise_variable_name, column_wise_variable_name],
        var_name='Value Type',
        value_name='Count'
    )
    
    # Show clustered bar chart of observed and expected values, if requested
    if show_plot:
        # Draw a nested barplot by species and sex
        ax = sns.catplot(
            data=df_counts,
            kind='bar',
            x=column_wise_variable_name,
            y='Count',
            hue='Value Type',
            col=row_wise_variable_name,
            palette=color_palette, 
            alpha=fill_transparency, 
            height=figure_size[1],
            aspect=figure_size[0] / figure_size[1],
            ci=None,
        )
        
        # Add space between the title and the plot
        plt.subplots_adjust(top=0.85)
        
        # Iterate through subplots
        for ax in ax.axes.flat:
            # Wrap x axis label using textwrap
            wrapped_x_labels = textwrap.fill(ax.get_xlabel(), 60, break_long_words=False)
            ax.set_xlabel(wrapped_x_labels, fontsize=10, fontname="Arial", color="#262626", horizontalalignment='center')
        
            # Format and wrap x axis tick labels using textwrap
            x_tick_labels = ax.get_xticklabels()
            wrapped_x_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 30, break_long_words=False)) for label in x_tick_labels]
            ax.set_xticklabels(wrapped_x_tick_labels, fontsize=10, fontname="Arial", color="#262626")
            
             # Word wrap the column labels without splitting words
            wrapped_column_labels = textwrap.fill(ax.get_title(), 60, break_long_words=False)
            ax.set_title(wrapped_column_labels, fontname="Arial", fontsize=10, color="#262626", horizontalalignment='center')
            
            # Change x-axis colors to "#666666"
            ax.tick_params(axis='x', colors="#666666")
            ax.spines['top'].set_color("#666666")
        
            # Remove bottom, left, and right spines
            ax.spines['bottom'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['left'].set_visible(False)
        
            # Remove the y-axis tick text
            ax.set_yticklabels([])
        
            # Add data labels to each bar in the plot
            # Get minimum value absolute value of the data
            min_value = df_counts['Count'].abs().min()
            if min_value < 1:
                min_value = 1
            
            # Iterate through each bar in the plot and add a data label
            for p in ax.patches:
                # Get the height of each bar
                height = p.get_height()
                
                # Format the height to the specified number of decimal places
                height = round(height, decimal_places_for_data_label)
                
                # Add the data label to the plot, using the specified number of decimal places
                value_label = "{:,.{decimal_places}f}".format(height, decimal_places=decimal_places_for_data_label)
                ax.text(
                    x=p.get_x() + p.get_width() / 2.,
                    y=height + (min_value * 0.05),
                    s=value_label,
                    ha="center",
                    fontname="Arial",
                    fontsize=10,
                    color="#262626"
                )
            
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=title_font_size,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=subtitle_font_size,
            color="#666666",
            transform=ax.transAxes
        )
        
            
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=caption_font_size,
                color="#666666",
                transform=ax.transAxes
            )
        
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
    
    
    # Pivot value type column before return obs vs exp counts
    df_counts = df_counts.pivot(
        index=[column_wise_variable_name, row_wise_variable_name],
        columns='Value Type',
        values='Count')
    df_counts = df_counts.rename_axis(columns = None).reset_index()
    
    # Add observed minus expected column
    df_counts['Difference (observed minus expected)'] = df_counts['Observed'] - df_counts['Expected']
    
    # Return counts 
    return(df_counts)




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import sympy

# Declare function
def FindDerivative(f_of_x,
                   print_functions=False,
                   return_derivative_function=False,
                   plot_functions=True,
                   minimum_x=-10, 
                   maximum_x=10, 
                   n=100):
    """
    This function finds the derivative of a function.

    Args:
        f_of_x (function): A function of x as a sympy expression.
        print_functions (bool, optional): Whether to print the function and its derivative. Defaults to True.
        return_derivative_function (bool, optional): Whether to return the derivative function. Defaults to False.

    Returns:
        d_f_of_x (function): The derivative of the function.
    """
    # Compute the derivative of the higher-order function using sympy
    try:
        d_f_of_x = sympy.diff(f_of_x, x)
    except NameError:
        raise ValueError("The function must be a sympy expression. Ensure that you have imported sympy and declared x as a sympy symbol. ( e.g., x = sympy.Symbol('x') )" )
    
    # Print the derivative function
    if print_functions:
        print("f(x):", f_of_x)
        print("f'(x):", d_f_of_x)
        
    # Plot the derivative function if requested
    if plot_functions:
        # Create x values
        x_values = np.linspace(minimum_x, maximum_x, n)
        
        # Vectorize the original function and its derivative
        vfunc = np.vectorize(lambda val: f_of_x.evalf(subs={x: val}))
        vfunc_derivative = np.vectorize(lambda val: d_f_of_x.evalf(subs={x: val}))

        # Create y values for original function and its derivative using the vectorized functions
        y_values_original = vfunc(x_values)
        y_values_derivative = vfunc_derivative(x_values)

        # Plot the original function
        plt.plot(x_values, y_values_original, label="f(x)")
        
        # Plot the derivative function
        plt.plot(x_values, y_values_derivative, label="f'(x)")
        
        # Add a legend
        plt.legend()
        
        # Show the plot
        plt.show()
    
    # Return the derivative function if requested
    if return_derivative_function:
        return d_f_of_x



# Load packages
import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotDensityByGroup(dataframe,
                       value_column_name,
                       grouping_column_name,
                       # Plot formatting arguments
                       color_palette='Set2',
                       show_legend=True,
                       # Text formatting arguments
                       title_for_plot=None,
                       subtitle_for_plot=None,
                       caption_for_plot=None,
                       data_source_for_plot=None,
                       show_y_axis=False,
                       title_y_indent=1.1,
                       subtitle_y_indent=1.05,
                       caption_y_indent=-0.15,
                       # Plot formatting arguments
                       figure_size=(8, 6)):
    """
    This function generates a density plot for a given dataframe, with the option to group by a specific column.
    The plot is generated using seaborn's kdeplot function.

    Args:
        dataframe (pandas.DataFrame): The dataframe to plot.
        value_column_name (str): The name of the column in the dataframe to use for the x-axis values.
        grouping_column_name (str): The name of the column in the dataframe to use for grouping data.
        color_palette (str, optional): The color palette to use for the plot. Defaults to 'Set2'.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).
    """
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
        
    # Generate density plot using seaborn
    sns.kdeplot(
        data=dataframe,
        x=value_column_name,
        hue=grouping_column_name,
        fill=True,
        common_norm=False,
        alpha=.4,
        palette=color_palette,
        linewidth=1
    )
    
    # Remove top, left, and right spines. Set bottom spine to dark gray.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_color("#262626")
    
    # Remove the legend if show_legend is False
    if show_legend == False:
        ax.legend_.remove()
    
    # Remove the y-axis, and adjust the indent of the plot titles
    if show_y_axis == False:
        ax.axes.get_yaxis().set_visible(False)
        x_indent = 0.015
    else:
        x_indent = -0.005
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # Set x-axis tick label font to Arial, size 9, and color #666666
    ax.tick_params(
        axis='x',
        which='major',
        labelsize=9,
        labelcolor="#666666",
        pad=2,
        bottom=True,
        labelbottom=True
    )
    # plt.xticks(fontname='Arial')
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages 
import numpy as np
import pandas as pd

# Declare function
def CreateRareCategoryColumn(dataframe,
                             categorical_column_name,
                             rare_category_label="Other",
                             rare_category_threshold=0.01,
                             new_column_suffix=None):
    """
    This function creates a new column in a dataframe with rare categories. 
    The default is to label rare categories as "Other". 
    The default threshold is 1%. If new_column_suffix is not specified, the new column name is the original column name with " (with Other)" appended to it.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        categorical_column_name (str): The name of the column containing the categorical variable.
        rare_category_label (str, optional): The label to use for rare categories. Defaults to "Other".
        rare_category_threshold (float, optional): The relative frequency threshold for rare categories. Defaults to 0.01.
        new_column_suffix (str, optional): The suffix to append to the original column name to create the new column name. Defaults to None.

    Returns:
        Pandas dataframe: An updated Pandas dataframe with the new column containing rare categories
    """
    
    # If new column name is not provided, create one
    if new_column_suffix is None:
        new_column_suffix = " (with " + rare_category_label + ")"
    
    # Get relative frequency of each category
    data_value_relative_frequency = dataframe[categorical_column_name].value_counts(normalize=True)
    data_value_relative_frequency = pd.DataFrame(data_value_relative_frequency)
    
    # Add flag for rare categories
    data_value_relative_frequency['rare'] = data_value_relative_frequency[categorical_column_name] < rare_category_threshold
    
    # Create new column name
    new_column_name = categorical_column_name + new_column_suffix
    
    # Get list of rare categories
    rare_values = data_value_relative_frequency[data_value_relative_frequency['rare'] == True].index.tolist()
    
    # Create new column in original dataframe with rare categories
    dataframe[new_column_name] = np.where(
        dataframe[categorical_column_name].isnull(), np.nan,
        np.where(dataframe[categorical_column_name].isin(rare_values), rare_category_label, 
        dataframe[categorical_column_name])
    )
    
    # Return the dataframe
    return dataframe




# Load packages
from scipy.stats import ttest_ind_from_stats
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def TTestOfTwoMeansFromStats(group1_mean,
                             group1_sd,
                             group1_sample_size,
                             group2_mean,
                             group2_sd,
                             group2_sample_size,
                             homogeneity_of_variance=True,
                             confidence_interval=0.95,
                             alternative_hypothesis="two-sided",
                             plot_sample_distributions=True,
                             value_name="Value",
                             color_palette='Set2',
                             title_for_plot="T-Test of Two Means",
                             subtitle_for_plot="Shows the distribution of the sample means for each group.",
                             caption_for_plot=None,
                             data_source_for_plot=None,
                             show_y_axis=False,
                             title_y_indent=1.1,
                             subtitle_y_indent=1.05,
                             caption_y_indent=-0.15,
                             figure_size=(8, 6)):
    """
    Conducts a two-sample t-test of the null hypothesis that the means of two independent samples are equal.

    Args:
        group1_mean (float): The mean of the first sample.
        group1_sd (float): The standard deviation of the first sample.
        group1_sample_size (int): The sample size of the first sample.
        group2_mean (float): The mean of the second sample.
        group2_sd (float): The standard deviation of the second sample.
        group2_sample_size (int): The sample size of the second sample.
        homogeneity_of_variance (bool, optional): Whether to assume equal variances for the two samples. Defaults to True.
        confidence_interval (float, optional): The confidence level for the test. Defaults to 0.95.
        alternative_hypothesis (str, optional): The alternative hypothesis for the test. Can be "two-sided", "less", or "greater". Defaults to "two-sided".
        plot_sample_distributions (bool, optional): Whether to plot the distribution of the sample means for each group. Defaults to True.
        value_name (str, optional): The name of the value being measured. Defaults to "Value".
        color_palette (str or list, optional): The color palette to use for the plot. Can be a string with the name of a seaborn palette or a list of colors. Defaults to 'Set2'.
        title_for_plot (str, optional): The title of the plot. Defaults to "T-Test of Two Means".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to "Shows the distribution of the sample means for each group.".
        caption_for_plot (str, optional): The caption of the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis of the plot. Defaults to False.
        title_y_indent (float, optional): The y-coordinate of the title of the plot. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-coordinate of the subtitle of the plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-coordinate of the caption of the plot. Defaults to -0.15.
        figure_size (tuple, optional): The size of the plot. Defaults to (8, 6).

    Returns:
        namedtuple: A named tuple containing the test statistic and the p-value of the test.
    """
    
    # Conduct t-test
    ttest_results = ttest_ind_from_stats(
        mean1=group1_mean, 
        std1=group1_sd, 
        nobs1=group1_sample_size,
        mean2=group2_mean, 
        std2=group2_sd, 
        nobs2=group2_sample_size,
        equal_var=homogeneity_of_variance,
        alternative=alternative_hypothesis
    )

    # Interpret test results
    if alternative_hypothesis == "two-sided" and ttest_results.pvalue <= (1-confidence_interval):
        print("There is a statistically detectable difference between the sample means of Group 1 and Group 2.")
    elif ttest_results.pvalue <= (1-confidence_interval):
        print("The sample mean of Group 1 is " + alternative_hypothesis + " than that of Group 2 to a statistically detectable extent.")
    else:
        print("Cannot reject the null hypothesis.")
    print("p-value:", str(round(ttest_results.pvalue, 3)))
    print("statistic:", str(round(ttest_results.statistic, 3)))
    
    # If plot_distribution_of_differences is True, plot distribution of differences
    if plot_sample_distributions:
        # Create a normal distribution for Group 1
        group1_distribution = np.random.normal(
            loc=group1_mean,
            scale=group1_sd / np.sqrt(group1_sample_size),  # Standard error
            size=10000
        )
        
        # Convert to dataframe
        group1_distribution = pd.DataFrame(group1_distribution,
                                           columns=[value_name])
        
        # Create a normal distribution for Group 2
        group2_distribution = np.random.normal(
            loc=group2_mean,
            scale=group2_sd / np.sqrt(group2_sample_size),  # Standard error
            size=10000
        )
        
        # Convert to dataframe
        group2_distribution = pd.DataFrame(group2_distribution,
                                           columns=[value_name])
        
        # Add group label
        group1_distribution["Group"] = "Group 1"
        group2_distribution["Group"] = "Group 2"
        
        # Row bind groups together
        data_distribution = pd.concat([group1_distribution, group2_distribution], axis=0).reset_index(drop=True)
        
        # Plot distribution of means
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
            
        # Generate density plot using seaborn
        sns.kdeplot(
            data=data_distribution,
            x=value_name,
            hue="Group",
            fill=True,
            common_norm=False,
            alpha=.4,
            palette=color_palette,
            linewidth=1
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
        
    # Reutrn results
    return ttest_results




# Load packages
import pandas as pd

# Declare function
def CountMissingDataByGroup(dataframe,
                            list_of_grouping_columns,
                            list_of_columns_to_analyze=None):
    """
    This function counts the number of records with missing data by group.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        list_of_grouping_columns (list): List of variables to group by
        list_of_columns_to_analyze (list, optional): List of variables to count missing data in. Defaults to None. If None, all variables are counted.

    Returns:
        Pandas dataframe: A Pandas dataframe with the number of records with missing data by group.
    """
    
    # Select variables to count missing data for
    if list_of_columns_to_analyze is not None:
        dataframe = dataframe[list_of_grouping_columns + list_of_columns_to_analyze]
    
    # Count missing data by group
    df_missing_by_group = dataframe.groupby(
        list_of_grouping_columns,
        dropna=False).apply(lambda x: x.isnull().sum()
    )
        
    # Get row count by group
    df_row_count_by_group = dataframe.groupby(
        list_of_grouping_columns,
        dropna=False
    ).size()
    df_missing_by_group['Row count'] = df_row_count_by_group
    del(df_row_count_by_group)
    
    # Set row count as first column
    first_column = df_missing_by_group.pop('Row count')
    df_missing_by_group.insert(0, 'Row count', first_column)
    
    # Return missing data by group summary
    return(df_missing_by_group)




# Load packages
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
import seaborn as sns
import textwrap

# Declare function
def PlotScatterplot(dataframe,
                    y_axis_column_name,
                    x_axis_column_name,
                    # Dot formatting arguments
                    grouping_column_name=None,
                    group_color_palette="Set1",
                    dot_fill_color="#999999",
                    size_by_column_name=None,
                    size_normalization=None,
                    # Fitted line arguments
                    fitted_line_type=None,
                    line_color=None,
                    # Text formatting arguments
                    title_for_plot=None,
                    subtitle_for_plot=None,
                    caption_for_plot=None,
                    data_source_for_plot=None,
                    x_indent=-0.128,
                    title_y_indent=1.125,
                    subtitle_y_indent=1.05,
                    caption_y_indent=-0.3,
                    # Quadrant arguments
                    upper_left_quadrant_label=None,
                    upper_left_quadrant_fill_color=None,
                    upper_right_quadrant_label=None,
                    upper_right_quadrant_fill_color=None,
                    lower_left_quadrant_label=None,
                    lower_left_quadrant_fill_color=None,
                    lower_right_quadrant_label=None,
                    lower_right_quadrant_fill_color=None,
                    folder_to_save_plot=None):
    """
    This function generates a scatterplot using the provided dataframe and column names for x and y axes.

    Args:
        dataframe (DataFrame): The dataframe containing the data.
        y_axis_column_name (str): The column name for the y-axis data.
        x_axis_column_name (str): The column name for the x-axis data.
        grouping_column_name (str, optional): The column name for grouping data. Defaults to None.
        group_color_palette (str, optional): The color palette for the groups. Defaults to "Set1".
        dot_fill_color (str, optional): The fill color for the dots. Defaults to "#999999".
        size_by_column_name(str, optional): The column name for sizing the dots. Defaults to None.
        size_normalization (tuple, optional): The range of sizes for the dots. Defaults to None.
        fitted_line_type (str, optional): The type of fitted line. Defaults to None.
        line_color (str, optional): The color of the line. Defaults to None.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x indent for the plot. Defaults to -0.128.
        title_y_indent (float, optional): The y indent for the title. Defaults to 1.125.
        subtitle_y_indent (float, optional): The y indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y indent for the caption. Defaults to -0.3.
        upper_left_quadrant_label (str, optional): The label for the upper left quadrant. Defaults to None.
        upper_left_quadrant_fill_color (str, optional): The fill color for the upper left quadrant. Defaults to None.
        upper_right_quadrant_label (str, optional): The label for the upper right quadrant. Defaults to None.
        upper_right_quadrant_fill_color (str, optional): The fill color for the upper right quadrant. Defaults to None.
        lower_left_quadrant_label (str, optional): The label for the lower left quadrant. Defaults to None.
        lower_left_quadrant_fill_color (str, optional): The fill color for the lower left quadrant. Defaults to None.
        lower_right_quadrant_label (str, optional): The label for the lower right quadrant. Defaults to None.
        lower_right_quadrant_fill_color (str, optional): The fill color for the lower right quadrant. Defaults to None.
        folder_to_save_plot (str, optional): The folder to save the plot. Defaults to None.
    """
    
    # Ensure that fitted_line_type is None, 'straight', or 'lowess'
    if fitted_line_type not in [None, 'straight', 'lowess']:
        raise ValueError("Invalid fitted_line_type argument. Please enter None, 'straight', or 'lowess'.")
    
    # If line color is not specified, use fill color
    if line_color == None:
        line_color = dot_fill_color
        
    # Create list of quadrant labels
    quadrant_labels = [upper_left_quadrant_label, upper_right_quadrant_label, lower_left_quadrant_label, lower_right_quadrant_label]
    
    # Draw scatterplot
    if fitted_line_type == None:
        if grouping_column_name == None:
            ax = sns.scatterplot(
                data=dataframe, 
                x=x_axis_column_name,
                y=y_axis_column_name,
                size=size_by_column_name,
                size_norm=size_normalization,
                color=dot_fill_color,
                alpha=0.5,
                linewidth=0.5,
                edgecolor=dot_fill_color
            )
        else:
            ax = sns.scatterplot(
                data=dataframe, 
                x=x_axis_column_name,
                y=y_axis_column_name,
                size=size_by_column_name,
                size_norm=size_normalization,
                hue=grouping_column_name,
                alpha=0.5,
                linewidth=0.5,
                palette=group_color_palette
            )
    elif fitted_line_type == 'straight':
        if grouping_column_name == None:
            ax = sns.regplot(
                data=dataframe,
                x=x_axis_column_name,
                y=y_axis_column_name,
                marker='o',
                scatter_kws={
                    'color': dot_fill_color,
                    'alpha': 0.5,
                    'linewidth': 0.5,
                    'edgecolor': dot_fill_color,
                    'size': size_by_column_name,
                    'size_norm': size_normalization,
                },
                linewidth=0.5,
                edgecolor=dot_fill_color,
                fit_reg=True,
                line_kws={'color': line_color}
            )
        else:
            ax = sns.lmplot(
                data=dataframe,
                x=x_axis_column_name,
                y=y_axis_column_name,
                hue=grouping_column_name,
                palette=group_color_palette,
                scatter_kws={
                    'alpha': 0.5,
                    'linewidth': 0.5,
                    'size': size_by_column_name,
                    'size_norm': size_normalization,
                },
                fit_reg=True
            )
    elif fitted_line_type == 'lowess':
        if grouping_column_name == None:
            ax = sns.regplot(
                data=dataframe,
                x=x_axis_column_name,
                y=y_axis_column_name,
                marker='o',
                scatter_kws={
                    'color': dot_fill_color,
                    'alpha': 0.5,
                    'linewidth': 0.5,
                    'edgecolor': dot_fill_color,
                    'size': size_by_column_name,
                    'size_norm': size_normalization,
                },
                lowess=True,
                line_kws={'color': line_color}
            )
        else:
            ax = sns.lmplot(
                data=dataframe,
                x=x_axis_column_name,
                y=y_axis_column_name,
                hue=grouping_column_name,
                palette=group_color_palette,
                scatter_kws={
                    'alpha': 0.5,
                    'linewidth': 0.5,
                    'size': size_by_column_name,
                    'size_norm': size_normalization,
                },
                lowess=True
            )
        
    # Remove top and right spines, and set bottom and left spines to gray
    try:
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
    except AttributeError:
        ax.ax.spines['top'].set_visible(False)
        ax.ax.spines['right'].set_visible(False)
        ax.ax.spines['bottom'].set_color('#666666')
        ax.ax.spines['left'].set_color('#666666')
    
    # Format tick labels to be Arial, size 9, and color #666666
    try:
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
    except AttributeError:
        ax.ax.tick_params(
            which='major', 
            labelsize=9, 
            color='#666666'
        )
    
    # If any of the quadrant labels are not None, find the middle of the x and y axes
    if any(quadrant_labels):
        try:
            x_start = ax.get_xlim()[0]
            x_end = ax.get_xlim()[1]
            y_start = ax.get_ylim()[0]
            y_end = ax.get_ylim()[1]
        except AttributeError:
            x_start = ax.ax.get_xlim()[0]
            x_end = ax.ax.get_xlim()[1]
            y_start = ax.ax.get_ylim()[0]
            y_end = ax.ax.get_ylim()[1]
        x_middle = x_start + (x_end - x_start) / 2
        y_middle = y_start + (y_end - y_start) / 2
        
        # Add upper left quadrant
        if upper_left_quadrant_label != None:
            # Get coordinates for upper left quadrant label
            upper_left_quadrant_label_x = x_start + (x_middle - x_start) / 2
            upper_left_quadrant_label_x = upper_left_quadrant_label_x + (x_start - upper_left_quadrant_label_x) / 2
            upper_left_quadrant_label_y = y_middle + (y_end - y_middle) / 2
            upper_left_quadrant_label_y = upper_left_quadrant_label_y + (y_end - upper_left_quadrant_label_y) / 2
            # If no fill color is specified, use dot fill color
            if upper_left_quadrant_fill_color == None:
                upper_left_quadrant_fill_color = dot_fill_color
            # Write text in upper left quadrant
            try:
                ax.text(
                    upper_left_quadrant_label_x, 
                    upper_left_quadrant_label_y, 
                    upper_left_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=upper_left_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            except AttributeError:
                ax.ax.text(
                    upper_left_quadrant_label_x, 
                    upper_left_quadrant_label_y, 
                    upper_left_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=upper_left_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            # Add a rectangle in the upper left quadrant, behind the dot plot
            if fitted_line_type == None:
                try:
                    ax.add_patch(
                        Rectangle(
                            (x_start, y_middle),
                            x_middle - x_start,
                            y_end - y_middle,
                            edgecolor='none',
                            facecolor=upper_left_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                except AttributeError:
                    ax.ax.add_patch(
                        Rectangle(
                            (x_start, y_middle),
                            x_middle - x_start,
                            y_end - y_middle,
                            edgecolor='none',
                            facecolor=upper_left_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
        
        # Add upper right quadrant
        if upper_right_quadrant_label != None:
            # Get coordinates for upper right quadrant label
            upper_right_quadrant_label_x = x_middle + (x_end - x_middle) / 2
            upper_right_quadrant_label_x = upper_right_quadrant_label_x + (x_end - upper_right_quadrant_label_x) / 2
            upper_right_quadrant_label_y = y_middle + (y_end - y_middle) / 2
            upper_right_quadrant_label_y = upper_right_quadrant_label_y + (y_end - upper_right_quadrant_label_y) / 2
            # If no fill color is specified, use dot fill color
            if upper_right_quadrant_fill_color == None:
                upper_right_quadrant_fill_color = dot_fill_color
            # Write text in upper right quadrant
            try:
                ax.text(
                    upper_right_quadrant_label_x, 
                    upper_right_quadrant_label_y, 
                    upper_right_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=upper_right_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            except AttributeError:
                ax.ax.text(
                    upper_right_quadrant_label_x, 
                    upper_right_quadrant_label_y, 
                    upper_right_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=upper_right_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            # Add a rectangle in the upper right quadrant, behind the dot plot
            if fitted_line_type == None:
                try:
                    ax.add_patch(
                        Rectangle(
                            (x_middle, y_middle),
                            x_end - x_middle,
                            y_end - y_middle,
                            edgecolor='none',
                            facecolor=upper_right_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                except AttributeError:
                    ax.ax.add_patch(
                        Rectangle(
                            (x_middle, y_end),
                            x_middle - x_end,
                            y_end - y_middle,
                            edgecolor='none',
                            facecolor=upper_right_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                
        # Add lower left quadrant
        if lower_left_quadrant_label != None:
            # Get coordinates for lower left quadrant label
            lower_left_quadrant_label_x = x_start + (x_middle - x_start) / 2
            lower_left_quadrant_label_x = lower_left_quadrant_label_x + (x_start - lower_left_quadrant_label_x) / 2
            lower_left_quadrant_label_y = y_start + (y_middle - y_start) / 2
            lower_left_quadrant_label_y = lower_left_quadrant_label_y - (y_middle - lower_left_quadrant_label_y) / 2
            # If no fill color is specified, use dot fill color
            if lower_left_quadrant_fill_color == None:
                lower_left_quadrant_fill_color = dot_fill_color
            # Write text in lower left quadrant
            try:
                ax.text(
                    lower_left_quadrant_label_x, 
                    lower_left_quadrant_label_y, 
                    lower_left_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=lower_left_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            except AttributeError:
                ax.ax.text(
                    lower_left_quadrant_label_x, 
                    lower_left_quadrant_label_y, 
                    lower_left_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=lower_left_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            # Add a rectangle in the lower left quadrant, behind the dot plot
            if fitted_line_type == None:
                try:
                    ax.add_patch(
                        Rectangle(
                            (x_start, y_start),
                            x_middle - x_start,
                            y_middle - y_start,
                            edgecolor='none',
                            facecolor=lower_left_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                except AttributeError:
                    ax.ax.add_patch(
                        Rectangle(
                            (x_start, y_start),
                            x_middle - x_start,
                            y_middle - y_start,
                            edgecolor='none',
                            facecolor=lower_left_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                
        # Add lower right quadrant
        if lower_right_quadrant_label != None:
            # Get coordinates for lower right quadrant label
            lower_right_quadrant_label_x = x_middle + (x_end - x_middle) / 2
            lower_right_quadrant_label_x = lower_right_quadrant_label_x + (x_end - lower_right_quadrant_label_x) / 2
            lower_right_quadrant_label_y = y_start + (y_middle - y_start) / 2
            lower_right_quadrant_label_y = lower_right_quadrant_label_y - (y_middle - lower_right_quadrant_label_y) / 2
            # If no fill color is specified, use dot fill color
            if lower_right_quadrant_fill_color == None:
                lower_right_quadrant_fill_color = dot_fill_color
            # Write text in lower right quadrant
            try:
                ax.text(
                    lower_right_quadrant_label_x, 
                    lower_right_quadrant_label_y, 
                    lower_right_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=lower_right_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            except AttributeError:
                ax.ax.text(
                    lower_right_quadrant_label_x, 
                    lower_right_quadrant_label_y, 
                    lower_right_quadrant_label, 
                    fontsize=10,
                    fontweight='bold',
                    color=lower_right_quadrant_fill_color,
                    alpha=0.8,
                    ha='center', 
                    va='center'
                )
            # Add a rectangle in the lower right quadrant, behind the dot plot
            if fitted_line_type == None:
                try:
                    ax.add_patch(
                        Rectangle(
                            (x_middle, y_start),
                            x_end - x_middle,
                            y_middle - y_start,
                            edgecolor='none',
                            facecolor=lower_right_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
                except AttributeError:
                    ax.ax.add_patch(
                        Rectangle(
                            (x_middle, y_start),
                            x_end - x_middle,
                            y_middle - y_start,
                            edgecolor='none',
                            facecolor=lower_right_quadrant_fill_color,
                            alpha=0.15
                        )
                    ).set_zorder(-1)
        
        # Insert a vertical line at the middle of the x axis
        plt.axvline(x_middle, color='#999999', linestyle='--', linewidth=1.5, alpha=0.5)
        
        # Insert a horizontal line at the middle of the y axis
        plt.axhline(y_middle, color='#999999', linestyle='--', linewidth=1.5, alpha=0.5)
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    try:
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
    except AttributeError:
        ax.ax.text(
            x=x_indent,
            y=title_y_indent-0.05,
            s=title_for_plot,
            # fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.ax.transAxes
        )
        
    # Set the subtitle with Arial font, size 11, and color #666666
    try:
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
    except AttributeError:
        ax.ax.text(
            x=x_indent,
            y=subtitle_y_indent-0.023,
            s=subtitle_for_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.ax.transAxes
        )
        
    # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
    try:
        ax.yaxis.set_label_coords(-0.1, 0.84)
        ax.yaxis.set_label_text(
            y_axis_column_name,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
    except AttributeError:
        ax.ax.yaxis.set_label_coords(-0.1, 0.87)
        ax.ax.yaxis.set_label_text(
            y_axis_column_name,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
    # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
    try:
        ax.xaxis.set_label_coords(0.9, -0.1)
        ax.xaxis.set_label_text(
            x_axis_column_name,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
    except AttributeError:
        ax.ax.xaxis.set_label_coords(0.9, -0.075)
        ax.ax.xaxis.set_label_text(
            x_axis_column_name,
            # fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        try:
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
        except AttributeError:
            ax.ax.text(
                x=x_indent,
                y=caption_y_indent+0.075,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.ax.transAxes
            )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
from IPython.display import display, HTML, Markdown
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler

# Delcare function
def ConductPrincipalComponentAnalysis(dataframe,
                                      list_of_numeric_columns=None,
                                      number_of_components=None,
                                      random_seed=412,
                                      display_pca_as_markdown=True):
    """
    Conducts principal component analysis on a given dataframe.

    Args:
        dataframe (pandas.DataFrame): The dataframe to perform PCA on.
        list_of_numeric_columns (list, optional): A list of column names to use for PCA. If not specified, all numeric columns will be used. Defaults to None.
        number_of_components (int, optional): The number of principal components to use. If not specified, the function will determine the optimal number of components. Defaults to None.
        random_seed (int, optional): The random seed to use for reproducibility. Defaults to 412.
        display_pca_as_markdown (bool, optional): Whether to display the principal components as a markdown table. Defaults to True.

    Returns:
        pandas.DataFrame: The original dataframe with the principal components added.
    """
    
    # If list_of_numeric_columns is not specified, then use all numeric variables
    if list_of_numeric_columns is None:
        list_of_numeric_columns = dataframe.select_dtypes(include=[np.number]).columns.tolist()
        
    # Select only numeric variables
    dataframe_pca = dataframe[list_of_numeric_columns].copy()
    
    # Remove missing values
    dataframe_pca = dataframe_pca.dropna()

    # Create list of column names
    list_of_column_names = dataframe_pca.columns.tolist()
        
    # Scale the data
    scaler = MinMaxScaler()
    dataframe_pca = pd.DataFrame(scaler.fit_transform(dataframe_pca))
    dataframe_pca.columns = list_of_column_names

    # Determine optimal number of components
    if number_of_components is None:
        model = PCA(random_state=random_seed).fit(dataframe_pca)
        plt.plot(model.explained_variance_ratio_)
        plt.ylabel('Explained Variance')
        plt.show()
        
        # Determine number of components to use
        for i in range(1, len(model.explained_variance_ratio_)):
            if sum(model.explained_variance_ratio_[0:i]) >= 0.90:
                number_of_components = i
                break
        print("Number of components to use: " + str(number_of_components))
        print("Total variance explained: " + str(round(sum(model.explained_variance_ratio_[0:i])*100, 2)) + "%")
        print("Be sure to review the scree plot to ensure that the number of components selected is appropriate for your purpose.")

    # Conduct PCA
    model = PCA(n_components=number_of_components,
                random_state=random_seed).fit(dataframe_pca)

    # Create dataframe of principal components
    data_pca = pd.DataFrame(
        model.components_,
        columns=dataframe_pca.columns
    )
    data_pca.index = ['PC' + str(i) for i in range(1, number_of_components+1)]

    # Print principal components
    print("\nPrincipal components:")
    if display_pca_as_markdown:
        display(data_pca)
    else:
        print(data_pca)

    # Add principal components to original dataframe
    for i in range(1, number_of_components+1):
        dataframe['PC' + str(i)] = data_pca.iloc[i-1, 0:len(dataframe.columns)].dot(dataframe_pca.transpose())

    # Return dataframe
    return(dataframe)




# Load packages
import pandas as pd

# Declare function
def CleanTextColumns(dataframe):
    """
    This function cleans string-type columns in a dataframe by removing leading and trailing spaces.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe
    
    Returns:
        Pandas dataframe: An updated Pandas dataframe with cleaned string-type columns.
    """
    
    # Iterate over columns, and clean string-type columns
    for col_name in dataframe.columns:
        if dataframe.dtypes[col_name] == 'O':
            dataframe[col_name] = dataframe[col_name].str.strip()
            
    # Return dataframe
    return(dataframe)




# Load packages
from math import ceil
from matplotlib import pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotBulletChart(dataframe, 
                    value_column_name, 
                    grouping_column_name,
                    target_value_column_name=None,
                    list_of_limit_columns=None,
                    value_maximum=None,
                    value_minimum=0,
                    # Bullet chart formatting arguments
                    display_order_list=None,
                    null_background_color='#dbdbdb',
                    background_color_palette=None,
                    background_alpha=0.8,
                    value_dot_size=200,
                    value_dot_color="#383838",
                    value_dot_outline_color="#ffffff",
                    value_dot_outline_width=2,
                    target_line_color='#bf3228',
                    target_line_width=4,
                    figure_size=(8, 6),
                    # Text formatting arguments
                    show_value_labels=True,
                    value_label_color="#262626",
                    value_label_format="{:.0f}",
                    value_label_font_size=12,
                    value_label_spacing=.65,
                    title_for_plot=None,
                    subtitle_for_plot=None,
                    caption_for_plot=None,
                    data_source_for_plot=None,
                    title_y_indent=1.15,
                    subtitle_y_indent=1.1,
                    caption_y_indent=-0.15):
    """
    This function plots a bullet chart using the specified dataframe, value column, and group column.
    The bullet chart is a variation of a bar chart that is used to display a single value in relation to a target value and a set of limit values.
    The function allows for customization of the bullet chart formatting and text formatting.
    
    Args:
        dataframe: pandas dataframe. The dataframe containing the data to be plotted.
        value_column_name: str. The name of the column in the dataframe containing the values to be plotted.
        grouping_column_name: str. The name of the column in the dataframe containing the groups to be plotted.
        target_value_column_name: str, default None. The name of the column in the dataframe containing the target values to be plotted.
        list_of_limit_columns: list of str, default None. A list of the names of the columns in the dataframe containing the limit values to be plotted.
        value_maximum: float, default None. The maximum value to be plotted on the chart. If None, the maximum value among the limit columns is used.
        value_minimum: float, default 0. The minimum value to be plotted on the chart.
        display_order_list: list of str, default None. A list of the groups in the dataframe in the order they should be plotted.
        null_background_color: str, default '#dbdbdb'. The color of the background area of the chart where there is no data.
        background_color_palette: str, default None. The color palette to be used for the limit columns. If None, a greyscale palette is used.
        background_alpha: float, default 0.8. The alpha value of the limit column colors.
        value_dot_size: int, default 200. The size of the dot representing the value on the chart.
        value_dot_color: str, default "#383838". The color of the dot representing the value on the chart.
        value_dot_outline_color: str, default "#ffffff". The color of the outline of the dot representing the value on the chart.
        value_dot_outline_width: int, default 2. The width of the outline of the dot representing the value on the chart.
        target_line_color: str, default '#bf3228'. The color of the line representing the target value on the chart.
        target_line_width: int, default 4. The width of the line representing the target value on the chart.
        figure_size: tuple of int, default (8, 6). The size of the figure containing the chart.
        show_value_labels: bool, default True. Whether or not to show data labels for the value on the chart.
        value_label_color: str, default "#262626". The color of the data labels for the value on the chart.
        value_label_format: str, default "{:.0f}". The format string for the data labels for the value on the chart.
        value_label_font_size: int, default 12. The font size of the data labels for the value on the chart.
        title_for_plot: str, default None. The title of the chart.
        subtitle_for_plot: str, default None. The subtitle of the chart.
        caption_for_plot: str, default None. The caption of the chart.
        data_source_for_plot: str, default None. The data source of the chart.
        title_y_indent: float, default 1.15. The y-indent of the title of the chart.
        subtitle_y_indent: float, default 1.1. The y-indent of the subtitle of the chart.
        caption_y_indent: float, default -0.15. The y-indent of the caption of the chart.
    """
    # Ensure that the number of unique values in the group column is equal to the number of rows in the dataframe
    if len(dataframe[grouping_column_name].unique()) != len(dataframe):
        raise ValueError("The number of unique values in the group column must be equal to the number of rows in the dataframe.")
    
    # If value_maximum is not specified, use the maximum value among the limit columns
    if value_maximum is None:
        if list_of_limit_columns is not None:
            value_maximum = dataframe[list_of_limit_columns].max().max()
        else:
            value_maximum = dataframe[value_column_name].max()
    
    # If display_order_list is provided, check that it contains all of the categories in the dataframe
    if display_order_list != None:
        if not set(display_order_list).issubset(set(dataframe[grouping_column_name].unique())):
            raise ValueError("display_order_list must contain all of the categories in the " + grouping_column_name + " column of dataframe.")
        else:
            # Reverse the display_order_list
            display_order_list = display_order_list[::-1]
            # Order the dataframe by the display_order_list
            dataframe[grouping_column_name] = pd.Categorical(dataframe[grouping_column_name], categories=display_order_list, ordered=True)
            dataframe = dataframe.sort_values(grouping_column_name)
    else:
        # If display_order_list is not provided, create one from the dataframe
        display_order_list = dataframe.sort_values(value_column_name, ascending=True)[grouping_column_name].unique()

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=figure_size)

    # Iterate over each row in the dataframe
    for index, row in dataframe.iterrows():
        group = row[grouping_column_name]
        value = row[value_column_name]
        if target_value_column_name != None:
            target = row[target_value_column_name]
            
        # Plot the background area color
        bars = ax.barh(
            y=group, 
            width=value_maximum-value_minimum, 
            color=null_background_color, 
            height=0.8
        )
        
        # Get the y-position of the background area color
        y_position = bars[0].get_y()
        
        # If limit columns are specified, plot the limit columns
        if list_of_limit_columns != None:
            # Get the colors from the specified color palette, using the number of limit columns
            if background_color_palette == None:
                # Create a greyscale palette
                colors = sns.light_palette("#acacad", len(list_of_limit_columns) + 1)
                # Reverse the palette
                colors = colors[::-1]
            else:
                colors = sns.color_palette(background_color_palette, len(list_of_limit_columns) + 1)

            # Rearrange the limit columns in descending order according to each of their values
            sorted_row = row[list_of_limit_columns].sort_values(ascending=False)
            
            # Iterate over each limit column
            for i in range(len(list_of_limit_columns)):
                # Plot the limit column value as a bar
                sorted_row_value = sorted_row[i]
                if i == 0 and np.isnan(sorted_row_value) == False:
                    ax.barh(
                        y=group, 
                        width=value_maximum-value_minimum, 
                        color=colors[i], 
                        height=0.8
                    )
                # If the limit column value is not null plot the limit column value as a bar:
                if np.isnan(sorted_row_value) == False:
                    ax.barh(
                        y=group, 
                        width=sorted_row_value-value_minimum, 
                        color=colors[i+1], 
                        height=0.8,
                        edgecolor='white',
                        linewidth=2,
                        alpha=background_alpha,
                    )
                else:
                    continue
        
        # Plot the target value as a line in the group
        if target_value_column_name != None:
            ax.plot(
                [target, target], 
                [y_position+0.2, y_position+0.6], 
                color=target_line_color, 
                linewidth=target_line_width
            )

        # Plot value as a large dot
        ax.scatter(
            value, 
            group, 
            color=value_dot_color, 
            s=value_dot_size,
            edgecolor=value_dot_outline_color,
            linewidth=value_dot_outline_width
        )
        
        # Add data label for the value, if requested
        if show_value_labels:
            ax.text(
                x=value,
                y=y_position+value_label_spacing,
                s=value_label_format.format(value),
                # fontname="Arial",
                fontsize=value_label_font_size,
                color=value_label_color,
                verticalalignment="center",
                horizontalalignment="center",
            )
        
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Format and wrap y axis tick labels using textwrap
    y_tick_labels = ax.get_yticklabels()
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in y_tick_labels]
    ax.set_yticklabels(wrapped_y_tick_labels, fontsize=10, fontname="Arial", color="#262626")
    
    # Move x-axis to the top
    ax.xaxis.tick_top()
    
    # Change x-axis colors to "#666666"
    ax.tick_params(axis='x', colors="#666666")
    ax.spines['top'].set_color("#666666")
    
    # Remove bottom, left, and right spines
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
        
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import ttest_ind
import seaborn as sns
import textwrap

# Declare function
def TwoSampleTTestOfIndependence(dataframe,
                                 outcome_column,
                                 grouping_column,
                                 confidence_interval=.95,
                                 alternative_hypothesis="two-sided",
                                 homogeneity_of_variance=True,
                                 null_handling='omit',
                                 plot_sample_distributions=True,
                                 color_palette='Set2',
                                 title_for_plot="T-Test of Independence",
                                 subtitle_for_plot="Shows the distribution of the sample means for each group.",
                                 caption_for_plot=None,
                                 data_source_for_plot=None,
                                 show_y_axis=False,
                                 title_y_indent=1.1,
                                 subtitle_y_indent=1.05,
                                 caption_y_indent=-0.15,
                                 figure_size=(8, 6)):
    """
    Conducts a two-sample t-test of independence on a dataframe with two groups.

    Args:
    - dataframe: pandas DataFrame containing the data to be analyzed
    - outcome_column: string representing the name of the column containing the outcome variable
    - grouping_column: string representing the name of the column containing the grouping variable
    - confidence_interval: float representing the desired confidence interval (default=.95)
    - alternative_hypothesis: string representing the alternative hypothesis (default="two-sided")
    - homogeneity_of_variance: boolean representing whether to assume equal variances (default=True)
    - null_handling: string representing how to handle null values (default='omit')
    - plot_sample_distributions: boolean representing whether to plot the distribution of the sample means for each group (default=True)
    - color_palette: string representing the name of the color palette to use for the plot (default='Set2')
    - title_for_plot: string representing the title of the plot (default="T-Test of Independence")
    - subtitle_for_plot: string representing the subtitle of the plot (default="Shows the distribution of the sample means for each group.")
    - caption_for_plot: string representing the caption of the plot (default=None)
    - data_source_for_plot: string representing the data source of the plot (default=None)
    - show_y_axis: boolean representing whether to show the y-axis on the plot (default=False)
    - title_y_indent: float representing the y-indent of the title on the plot (default=1.1)
    - subtitle_y_indent: float representing the y-indent of the subtitle on the plot (default=1.05)
    - caption_y_indent: float representing the y-indent of the caption on the plot (default=-0.15)
    - figure_size: tuple representing the size of the plot (default=(8, 6))

    Returns:
    - ttest_results: a scipy.stats.ttest_ind_result object containing the results of the t-test
    """
    
    # Get groupings (there should only be two)
    list_groups = dataframe[grouping_column].unique()
    if len(list_groups) > 2:
        raise ValueError("T-Test can only be performed with two groups. The grouping column you selected has " + str(len(list_groups)) + " groups.")
    
    # Show pivot table summary
    table_groups = pd.pivot_table(dataframe, 
        values=[outcome_column], 
        index=[grouping_column],
        aggfunc={outcome_column: [len, min, max, np.mean, np.std]}
    )
    print(table_groups)
    print("\n")
    
    # Restructure data
    dataframe = dataframe[[
        outcome_column,
        grouping_column,
    ]]
    group1 = dataframe[dataframe[grouping_column]==list_groups[0]]
    group2 = dataframe[dataframe[grouping_column]==list_groups[1]]
    
    # Conduct two sample t-test
    ttest_results = ttest_ind(
        group1[outcome_column], 
        group2[outcome_column], 
        axis=0,
        nan_policy=null_handling, 
        alternative=alternative_hypothesis,
        equal_var=homogeneity_of_variance
    )
    
    # Get the p-value
    p_value = ttest_results.pvalue
    
    # Get the test statistic
    test_statistic = ttest_results.statistic
    
    # If plot_distribution_of_differences is True, plot distribution of differences
    if plot_sample_distributions:
        # Create placeholder for results
        data_distribution = pd.DataFrame()
        
        # Get unqiue groups
        list_of_groups = dataframe[grouping_column].unique()
        
        # Iterate through groups
        for group in list_of_groups:
            # Get the group's data
            group_data = dataframe[dataframe[grouping_column] == group][outcome_column]
            
            # Get the group's mean
            group_mean = group_data.mean()
            
            # Get the group's standard error
            group_se = group_data.sem()
            
            # Generate a normal distribution
            group_distribution = np.random.normal(group_mean, group_se, 10000)
            
            # Convert to dataframe
            df_group_distribution = pd.DataFrame(group_distribution,
                                                 columns=[outcome_column])
            
            # Add group name
            df_group_distribution[grouping_column] = group
            
            # Row bind to data_distribution
            data_distribution = pd.concat([data_distribution, df_group_distribution], axis=0).reset_index(drop=True)
            
        # Plot distribution of means
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
            
        # Generate density plot using seaborn
        sns.kdeplot(
            data=data_distribution,
            x=outcome_column,
            hue=grouping_column,
            fill=True,
            common_norm=False,
            alpha=.4,
            palette=color_palette,
            linewidth=1
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()

    # Print interpretation of results
    if alternative_hypothesis == "two-sided" and ttest_results.pvalue <= (1-confidence_interval):
        print("There is a statistically detectable difference between the sample means of the groups.")
    elif ttest_results.pvalue <= (1-confidence_interval):
        print("The sample mean of the " + str(list_groups[0]) + " group is " + alternative_hypothesis + " than that of the " + str(list_groups[1]) + " group to a statistically detectable extent.")
    else:
        print("Cannot reject the null hypothesis.")
    print("p-value:", str(round(p_value, 3)))
    print("statistic:", str(round(test_statistic, 3)))
    
    # Return results
    return ttest_results




# Load packages
from fuzzywuzzy import fuzz
import Levenshtein
import numpy as np
import pandas as pd

# Declare function
def ConductEntityMatching(dataframe_1, 
                          dataframe_1_primary_key, 
                          dataframe_2, 
                          dataframe_2_primary_key,
                          levenshtein_distance_filter=None,
                          match_score_threshold=95,
                          columns_to_compare=None,
                          match_methods=['Partial Token Set Ratio', 'Weighted Ratio']):
    """
    Conducts entity matching between two dataframes using various fuzzy matching algorithms.
    
    Args:
        dataframe_1 (pandas.DataFrame): First dataframe to match.
        dataframe_1_primary_key (str): Name of the primary key column in dataframe_1.
        dataframe_2 (pandas.DataFrame): Second dataframe to match.
        dataframe_2_primary_key (str): Name of the primary key column in dataframe_2.
        levenshtein_distance_filter (int, optional): Maximum Levenshtein distance between two entities to consider a match. Defaults to None.
        match_score_threshold (int, optional): Minimum match score to consider a match. Defaults to 95.
        columns_to_compare (list of str, optional): List of columns to compare between the two dataframes. If None, all columns are compared. Defaults to None.
        match_methods (list of str, optional): List of fuzzy matching algorithms to use. Defaults to ['Partial Token Set Ratio', 'Weighted Ratio'].
        
    Returns:
        pandas.DataFrame: Dataframe of matched entities and their match scores.
    """
    
    # Create list of valid match methods
    valid_match_methods = [
        'Ratio', 'Partial Ratio', 
        'Token Sort Ratio', 'Partial Token Sort Ratio', 
        'Token Set Ratio', 'Partial Token Set Ratio', 
        'Weighted Ratio'
    ]
    
    # Ensure that match methods are valid
    for match_method in match_methods:
        if match_method not in valid_match_methods:
            raise ValueError('Invalid match method specified: ' + match_method + ". Valid match methods are: " + ', '.join(valid_match_methods))
        
    # Ensure that lev distance threshold is either int or None
    if levenshtein_distance_filter is not None:
        if not isinstance(levenshtein_distance_filter, int):
            raise ValueError('Levenshtein distance threshold must be an integer or None')

    # Select columns to compare if specified
    if columns_to_compare is not None:
        dataframe_1 = dataframe_1[[dataframe_1_primary_key] + columns_to_compare].copy()
        dataframe_2 = dataframe_2[[dataframe_2_primary_key] + columns_to_compare].copy()
        
    # Remove duplicate rows from dataframe_1 and dataframe_2
    dataframe_1.drop_duplicates(inplace=True)
    dataframe_2.drop_duplicates(inplace=True)
    
    # Create placeholder dataframe to store results
    data_match_results = pd.DataFrame()
    
    # Add string column to dataframe_1 to store concatenated values of columns to compare
    if len(columns_to_compare) == 1:
        dataframe_1['Entity'] = dataframe_1[columns_to_compare[0]]
        dataframe_2['Entity'] = dataframe_2[columns_to_compare[0]]
    else:
        dataframe_1['Entity'] = dataframe_1[columns_to_compare[0]].str.cat(dataframe_1[columns_to_compare[1:]], sep=' ')
        dataframe_2['Entity'] = dataframe_2[columns_to_compare[0]].str.cat(dataframe_2[columns_to_compare[1:]], sep=' ')
    
    # Create combination of each Entity 1 and Entity 2
    entity_combinations = pd.merge(dataframe_1, dataframe_2, how='cross', suffixes=('_1', '_2'))
    
    # Calculate Levenshtein distance if specified
    if levenshtein_distance_filter is not None:
        entity_combinations['Levenshtein Distance'] = entity_combinations.apply(lambda row: Levenshtein.distance(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        entity_combinations = entity_combinations[entity_combinations['Levenshtein Distance'] <= levenshtein_distance_filter]
    
    # If dataframe_1_primary_key is not in entity_combinations, add _1 to the end of the column name
    if dataframe_1_primary_key not in entity_combinations.columns:
        dataframe_1_primary_key = dataframe_1_primary_key + '_1'
    
    # If dataframe_2_primary_key is not in entity_combinations, add _2 to the end of the column name
    if dataframe_2_primary_key not in entity_combinations.columns:
        dataframe_2_primary_key = dataframe_2_primary_key + '_2'
    
    # Create list of columns to select
    list_columns_to_select = [
        'Entity_1', 
        'Entity_2'
    ]
    
    # Calculate fuzz ratio, which calls difflib.ratio on the two input strings
    if 'Ratio' in match_methods:
        # Calculate fuzz ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
    
    # Calculate partial ratio, which calls ratio using the shortest string (length n) against all n-length substrings of the larger string and returns the highest score
    if 'Partial Ratio' in match_methods:
        # Calculate partial fuzz ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Partial Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.partial_ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
        
    # Calculate token sort ratio, which tokenizes the strings and sorts them alphabetically before computing ratio
    if 'Token Sort Ratio' in match_methods:
        # Calculate token sort ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Token Sort Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.token_sort_ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
        
    # Calculate partial token sort ratio, which tokenizes the strings and sorts them alphabetically before computing ratio using the shortest string (length n) against all n-length substrings of the larger string and returns the highest score
    if 'Partial Token Sort Ratio' in match_methods:
        # Calculate partial token sort ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Partial Token Sort Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.partial_token_sort_ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
        
    # Calculate token set ratio, which tokenizes the strings and computes ratio using the intersection of the two token sets
    if 'Token Set Ratio' in match_methods:
        # Calculate token set ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Token Set Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.token_set_ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
        
    # Calculate partial token set ratio, which tokenizes the strings and computes ratio using the intersection of the two token sets using the shortest string (length n) against all n-length substrings of the larger string and returns the highest score
    if 'Partial Token Set Ratio' in match_methods:
        # Calculate partial token set ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Partial Token Set Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.partial_token_set_ratio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
    
    # Calculate weighted ratio, which is a weighted average of the above scores
    if 'Weighted Ratio' in match_methods:
        # Calculate weighted ratio
        df_temp = entity_combinations[list_columns_to_select].copy()
        df_temp['Match Method'] = 'Weighted Ratio'
        df_temp['Match Score'] = entity_combinations.apply(lambda row: fuzz.WRatio(row['Entity_1'], row['Entity_2']), axis=1)
        # Filter out combinations that are not within the threshold
        df_temp = df_temp[df_temp['Match Score'] >= match_score_threshold]
        # Add the matches to the dataframe
        data_match_results = pd.concat([data_match_results, df_temp], ignore_index=True)
    
    # Drop duplicates
    data_match_results.drop_duplicates(inplace=True)
    
    # Convert to wide-form dataframe
    data_match_results = data_match_results.pivot(index=['Entity_1', 'Entity_2'], columns='Match Method', values='Match Score')
    # Reset index of dataframe
    data_match_results['Entity 1'] = data_match_results.index.get_level_values(0)
    data_match_results['Entity 2'] = data_match_results.index.get_level_values(1)
    data_match_results.reset_index(drop=True, inplace=True)
    
    # Move entity columns to front
    data_match_results = data_match_results[['Entity 1', 'Entity 2'] + [col for col in data_match_results.columns if col not in ['Entity 1', 'Entity 2']]]
    
     # Reset column names
    data_match_results.columns.name = None
    
    # Left join columns to compare from dataframe 1
    if dataframe_1_primary_key not in dataframe_1.columns:
        dataframe_1_primary_key = dataframe_1_primary_key.replace('_1', '')
    data_match_results = data_match_results.merge(
        dataframe_1[[dataframe_1_primary_key, 'Entity'] + columns_to_compare], 
        how='left', 
        left_on='Entity 1', 
        right_on='Entity',
    )
    
    # Left join columns to compare from dataframe 2
    if dataframe_2_primary_key not in dataframe_2.columns:
        dataframe_2_primary_key = dataframe_2_primary_key.replace('_2', '')
    data_match_results = data_match_results.merge(
        dataframe_2[[dataframe_2_primary_key, 'Entity'] + columns_to_compare],
        how='left',
        left_on='Entity 2',
        right_on='Entity',
        suffixes=('_Entity 1', '_Entity 2')
    )
    
    # Drop entity columns
    data_match_results.drop(columns=['Entity_Entity 1', 'Entity_Entity 2'], inplace=True)
    
    # Move identifier columns to front
    if dataframe_1_primary_key not in data_match_results.columns:
        dataframe_1_primary_key = dataframe_1_primary_key + '_Entity 1'
    if dataframe_2_primary_key not in data_match_results.columns:
        dataframe_2_primary_key = dataframe_2_primary_key + '_Entity 2'
    data_match_results = data_match_results[[dataframe_1_primary_key, dataframe_2_primary_key] + [col for col in data_match_results.columns if col not in [dataframe_1_primary_key, dataframe_2_primary_key]]]
    
    # Drop duplicate matches
    data_match_results.drop_duplicates(inplace=True)
    
    # Return results
    return(data_match_results)




# Load packages 
from scipy import stats
from math import sqrt
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def TTestOfMeanFromStats(sample_mean,
                         sample_sd,
                         sample_size,
                         hypothesized_mean,
                         alternative_hypothesis="two-sided",
                         confidence_interval=.95,
                         plot_sample_distribution=True,
                         value_name="Value",
                         fill_color="#999999",
                         fill_transparency=0.6,
                         title_for_plot="Hypothesis Test of a Mean",
                         subtitle_for_plot="Shows the distribution of the sample mean and the hypothesized mean.",
                         caption_for_plot=None,
                         data_source_for_plot=None,
                         show_y_axis=False,
                         title_y_indent=1.10,
                         subtitle_y_indent=1.05,
                         caption_y_indent=-0.15,
                         figure_size=(8, 6)):
    """
    Conducts a hypothesis test of a mean using a t-test or z-test, depending on the sample size.

    Args:
        sample_mean (float): The sample mean.
        sample_sd (float): The sample standard deviation.
        sample_size (int): The sample size.
        hypothesized_mean (float): The hypothesized mean.
        alternative_hypothesis (str, optional): The alternative hypothesis. Can be "two-sided", "less", or "greater".
            Defaults to "two-sided".
        confidence_interval (float, optional): The confidence level for the hypothesis test. Defaults to 0.95.
        plot_sample_distribution (bool, optional): Whether to plot the distribution of the sample mean and the
            hypothesized mean. Defaults to True.
        value_name (str, optional): The name of the value being tested. Defaults to "Value".
        fill_color (str, optional): The fill color for the histogram. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the fill color for the histogram. Defaults to 0.6.
        title_for_plot (str, optional): The title for the plot. Defaults to "Hypothesis Test of a Mean".
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to "Shows the distribution of the sample
            mean and the hypothesized mean.".
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.10.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).

    Returns:
        float: The test statistic.
    """
    
    # If alternative hypothesis is "two-sided", then divide and add half of complement
    if alternative_hypothesis == "two-sided":
        ppf_threshold = confidence_interval + ((1 - confidence_interval) / 2)
    else:
        ppf_threshold = confidence_interval

    # If sample size 30 or more, calculate z-stat. Otherwise, calculate t-stat
    if sample_size < 30:
        test_threshold = stats.t.ppf(ppf_threshold)
    else:
        test_threshold = stats.norm.ppf(ppf_threshold)

    # Calculate difference between sample and hypothesized mean
    diff_in_means = hypothesized_mean - sample_mean

    # Calculate standard error of sample
    standard_error = sample_sd / sqrt(sample_size)

    # Calculate test statistic
    test_stat = diff_in_means / standard_error

    # Interpret results
    if alternative_hypothesis == "two-sided" and abs(test_stat) > test_threshold:
        print("There is a statistically detectable difference between the hypothesized mean and the sample mean.")
        if sample_size < 30:
            p_value = stats.t.cdf(abs(test_stat)*-1)*2
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
        else:
            p_value = stats.norm.cdf(abs(test_stat)*-1)*2
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
    elif alternative_hypothesis == "less" and test_stat < test_threshold*-1:
        print("The hypothesized mean is", alternative_hypothesis, "than the sample mean to a statistically detectable extent.")
        if sample_size < 30:
            p_value = stats.t.cdf(abs(test_stat)*-1)
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
        else:
            p_value = stats.norm.cdf(abs(test_stat)*-1)
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
    elif alternative_hypothesis == "greater" and test_stat > test_threshold:
        print("The hypothesized mean is", alternative_hypothesis, "than the sample mean to a statistically detectable extent.")
        if sample_size < 30:
            p_value = 1 - stats.t.cdf(abs(test_stat))
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
        else:
            p_value = 1 - stats.norm.cdf(abs(test_stat))
            print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
    else:
        print("Cannot reject the null hypothesis")
        if sample_size < 30:
            if alternative_hypothesis != "greater":
                p_value = stats.t.cdf(test_stat)
                print("p-value:", str(round(p_value, 3)))
            else:
                p_value = 1 - stats.t.cdf(test_stat)
                print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
        else:
            if alternative_hypothesis != "greater":
                p_value = stats.norm.cdf(test_stat)
                print("p-value:", str(round(p_value, 3)))
            else:
                p_value = 1 - stats.norm.cdf(test_stat)
                print("p-value:", str(round(p_value, 3)))
            print("statistic:", str(round(test_stat, 3)))
            
    # Generate plot of distribution of sample mean and hypothesized mean
    if plot_sample_distribution:
        # Generate an array of sample means
        sample_means = np.random.normal(
            loc=sample_mean, 
            scale=standard_error, 
            size=10000
        )
        
        # Convert array to dataframe
        sample_means = pd.DataFrame(sample_means,
                                    columns=[value_name])
        
        # # Flag values if they are outside of the confidence interval
        # if alternative_hypothesis=="two-sided":
        #     sample_means['Null Rejection Zone'] = np.where(
        #         sample_means[value_name] < hypothesized_mean, True,
        #         np.where(sample_means[value_name] > hypothesized_mean, True,
        #         False
        #     ))
        # elif alternative_hypothesis=="less":
        #     sample_means['Null Rejection Zone'] = np.where(
        #         sample_means[value_name] < hypothesized_mean, True,
        #         False
        #     )
        # elif alternative_hypothesis=="greater":
        #     sample_means['Null Rejection Zone'] = np.where(
        #         sample_means[value_name] > hypothesized_mean, True,
        #         False
        #     )
        
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        ax = sns.histplot(
            data=sample_means,
            x=value_name,
            # hue='Null Rejection Zone',
            # palette={True: "#FF0000", False: fill_color},
            alpha=fill_transparency,
            color=fill_color,
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the hypothesis value as a vertical line with a label
        ax.axvline(
            x=hypothesized_mean,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=hypothesized_mean, 
            y=plt.ylim()[1] * 0.97, 
            s='Hypothesized value: {:.2f}'.format(hypothesized_mean)+' ({:.1%})'.format(p_value),
            horizontalalignment='center',
            fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()

    # Return results
    return test_stat




# Load packages
import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotContingencyHeatmap(dataframe,
                           categorical_column_name_1,
                           categorical_column_name_2,
                           # Plot formatting arguments
                           color_palette="Blues",
                           show_legend=False,
                           figure_size=(8, 6),
                           # Text formatting arguments
                           data_label_format=".1%",
                           title_for_plot=None,
                           subtitle_for_plot=None,
                           caption_for_plot=None,
                           data_source_for_plot=None,
                           title_y_indent=1.15,
                           subtitle_y_indent=1.1,
                           caption_y_indent=-0.15):
    """
    This function generates a heatmap plot for two categorical variables from a given dataframe.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data.
        categorical_column_name_1 (str): The name of the first categorical column in the dataframe.
        categorical_column_name_2 (str): The name of the second categorical column in the dataframe.
        color_palette (str, optional): The color palette to use for the heatmap. Defaults to "Blues".
        show_legend (bool, optional): Whether to show the legend on the plot. Defaults to False.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).
        data_label_format (str, optional): The format for the data labels. Defaults to ".1%".
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.15.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.1.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
    """
    
    # Create contingency table
    contingency_table = pd.crosstab(
        dataframe[categorical_column_name_1],
        dataframe[categorical_column_name_2],
        normalize="columns"
    )
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Plot contingency table in a heatmap using seaborn
    ax = sns.heatmap(
        contingency_table,
        cmap=color_palette,
        annot=True,
        fmt=data_label_format,
        linewidths=0.5,
        cbar=show_legend
    )
    
    # Wrap y axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(categorical_column_name_1, 30))  # String wrap the variable name
    ax.set_ylabel(wrapped_variable_name)
    
    # Format and wrap y axis tick labels using textwrap
    y_tick_labels = ax.get_yticklabels()
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in y_tick_labels]
    ax.set_yticklabels(
        wrapped_y_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Wrap x axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(categorical_column_name_2, 30))  # String wrap the variable name
    ax.set_xlabel(wrapped_variable_name)
    
    # Format and wrap x axis tick labels using textwrap
    x_tick_labels = ax.get_xticklabels()
    wrapped_x_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in x_tick_labels]
    ax.set_xticklabels(
        wrapped_x_tick_labels, 
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
        
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
import pandas as pd
import numpy as np

# Declare function
def AddTPeriodColumn(dataframe,
                     date_column_name,
                     t_period_interval="days",
                     t_period_column_name=None):
    """
    This function adds a T-period column to a dataframe. 
    The T-period column is the number of intervals (e.g., days or weeks) since the earliest date in the dataframe.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        date_column_name (str): The name of the column containing the date. Options are "days", "weeks", "months", or "years"
        t_period_interval (str, optional): The interval or time object to use for the T-period column. Defaults to "days".
        t_period_column_name (str, optional): The name of the T-period column. Defaults to None. If None, the column name is "T Period in " + t_period_interval.

    Returns:
        Pandas dataframe: An updated Pandas dataframe with a T-period column.
    """
    
    # Ensure that column is a date datatype
    if dataframe[date_column_name].dtypes != "<M8[ns]":
        dataframe[date_column_name] = pd.to_datetime(dataframe[date_column_name])
        
    # Set T-period column name
    if t_period_column_name == None:
        t_period_column_name = "T Period in " + t_period_interval

    # Calculate difference from earliest date in interval specified
    earliest_time = min(dataframe[date_column_name])
    dataframe[t_period_column_name] = dataframe[date_column_name] - earliest_time
    if t_period_interval == "days":
        dataframe[t_period_column_name] = dataframe[t_period_column_name].dt.days
    if t_period_interval == "weeks":
        dataframe[t_period_column_name] = dataframe[t_period_column_name] / np.timedelta64(1, 'W')
        dataframe[t_period_column_name] = dataframe[t_period_column_name].apply(np.floor)
    if t_period_interval == "months":
        dataframe[t_period_column_name] = dataframe[t_period_column_name] / np.timedelta64(1, 'M')
        dataframe[t_period_column_name] = dataframe[t_period_column_name].apply(np.floor)
    if t_period_interval == "years":
        dataframe[t_period_column_name] = dataframe[t_period_column_name] / np.timedelta64(1, 'Y')
        dataframe[t_period_column_name] = dataframe[t_period_column_name].round(0)
    
    # Return updated dataframe
    return(dataframe)



# Description: Setup file for AnalysisToolbox package.
from setuptools import setup, find_packages

# Read the README file.
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

# Setup the package.
setup(
    name='analysistoolbox',
    description='A collection tools in Python for data collection and processing, statisitics, analytics, and intelligence analysis.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/KyleProtho/AnalysisToolBox/tree/master/analysistoolbox",
    version='1.3.8',
    author='Kyle Protho',
    author_email='kyletprotho@gmail.com',
    license='GNU GPLv3',
    classifiers=[
        # "Development Status :: 1 - Alpha",
        # "Intended Audience :: Analysts",
        # "Topic :: Intelligence Analysis, Data Science, Statistics",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    # package_dir={'': 'analysistoolbox'},
    packages=find_packages(),
    install_requires=[
        'beautifulsoup4',
        'folium',
        'fuzzywuzzy',
        'geopandas',
        'Jinja2',
        'Levenshtein',
        'lida',
        'lifelines',
        'mapclassify',
        'matplotlib',
        'mlxtend',
        'numpy',
        'openai',
        'pandas',
        'pygris',
        'pymetalog',
        'PyPDF2',
        'python-dotenv',
        'pywin32',
        'requests',
        'scipy',
        'seaborn',
        'scikit-learn',  # sklearn
        'statsmodels',
        'sympy',
        'tableone',
        'tensorflow',
        'xgboost',
        'yellowbrick'
    ],
    extras_require={
        'dev': [
            'pytest>=3.7',
            'twine>=4.0.2'
        ],
    },
)



# Load packages
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
import statsmodels.api as sm

# Declare function
def ConductLinearRegressionAnalysis(dataframe,
                                    outcome_variable,
                                    list_of_predictors,
                                    scale_predictors=False,
                                    show_diagnostic_plots_for_each_predictor=False,
                                    show_help=True):
    """
    Conducts a linear regression analysis on a given dataframe, using a specified outcome variable and list of predictor variables.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be analyzed.
        outcome_variable (str): The name of the column in the dataframe containing the outcome variable.
        list_of_predictors (list): A list of strings, each containing the name of a column in the dataframe containing a predictor variable.
        scale_predictors (bool, optional): Whether or not to scale the predictor variables before running the analysis. Defaults to False.
        show_diagnostic_plots_for_each_predictor (bool, optional): Whether or not to show diagnostic plots for each predictor variable. Defaults to False.
        show_help (bool, optional): Whether or not to show help text explaining how to access the output of the function. Defaults to True.

    Returns:
        dict: A dictionary containing the fitted linear regression model, the model summary, a test dataset of predictors, and a test dataset of outcomes.
    """
    
    # Select columns specified
    dataframe = dataframe[list_of_predictors + [outcome_variable]]
    
    # Remove NAN and inf values
    dataframe.dropna(inplace=True)
    dataframe = dataframe[np.isfinite(dataframe).all(1)]

    # Add constant
    dataframe = sm.add_constant(dataframe)
    
    # Scale the predictors, if requested
    if scale_predictors:
        # Show the mean and standard deviation of each predictor
        print("\nMean of each predictor:")
        print(dataframe[list_of_predictors].mean())
        print("\nStandard deviation of each predictor:")
        print(dataframe[list_of_predictors].std())
        
        # Scale predictors
        dataframe[list_of_predictors] = StandardScaler().fit_transform(dataframe[list_of_predictors])
    
    # Create linear regression model
    model = sm.OLS(dataframe[outcome_variable], dataframe[['const'] + list_of_predictors])
    model_res = model.fit()
    model_summary = model_res.summary()
    
    # If requested, show diagnostic plots
    if show_diagnostic_plots_for_each_predictor:
        for variable in list_of_predictors:
            fig = plt.figure(figsize=(12, 8))
            fig = sm.graphics.plot_regress_exog(model_res, variable, fig=fig)
    
    # If requested, show help text
    if show_help:
        print(
            "Quick guide on accessing output of ConductLinearRegressionAnalysis function:",
            "\nThe ouput of the ConductLinearRegressionAnalysis function is a dictionary containing the regression results, a test dataset of predictors, and a test dataset of outcomes.",
            "\n\t--To access the linear regression model, use the 'Fitted Model' key."
            "\n\t--To view the model's statistical summary, use the 'Model Summary' key."
        )
    
    # Create dictionary of objects to return
    dict_return = {
        "Fitted Model": model_res,
        "Model Summary": model_summary
    }
    return dict_return




# Load packages
import requests
import PyPDF2

# Declare function
def FetchPDFFromURL(url, filename):
    """This function downloads a PDF from a website and saves it to the current directory.

    Args:
        url (str): The URL of the PDF file that you want to download.
        filename (str): The name of the file that you want to save the PDF to.
    """
    
    # Send a GET request to the url and get the response
    response = requests.get(url)
    
    # Check if the response status code is 200 (OK)
    if response.status_code == 200:
        # Open the filename in write-binary mode
        with open(filename, "wb") as file:
            # Write the response content to the file
            file.write(response.content)
            # Print a success message
            print(f"PDF downloaded and saved as {filename}")
    else:
        # Print an error message
        print(f"Failed to download PDF from {url}")




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.stattools import adfuller
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
import textwrap
import warnings

# Declare function
def CreateARIMAModel(dataframe,
                     outcome_column_name,
                     time_column_name=None,
                     # Model parameter arguments
                     lookback_periods=1,
                     differencing_periods=0,
                     lag_periods=1,
                     moving_average_periods=1,
                     # Output arguments
                     plot_time_series=False,
                     time_series_figure_size=(8, 5),
                     # Line formatting arguments
                     line_color="#3269a8",
                     line_alpha=0.8,
                     # Text formatting arguments
                     number_of_x_axis_ticks=None,
                     x_axis_tick_rotation=None,
                     title_for_plot="Time Series",
                     subtitle_for_plot="Shows the time series for the outcome variable.",
                     caption_for_plot=None,
                     data_source_for_plot=None,
                     x_indent=-0.127,
                     title_y_indent=1.125,
                     subtitle_y_indent=1.05,
                     caption_y_indent=-0.3,
                     # ARIMA plot arguments
                     test_for_stationarity=True,
                     show_acf_pacf_plots=False,
                     show_model_results=False,
                     plot_residuals=True):
    """
    Creates an ARIMA model for time series analysis.

    Args:
        dataframe (pandas.DataFrame): The input dataframe containing the time series data.
        outcome_column_name (str): The name of the column in the dataframe representing the outcome variable.
        time_column_name (str, optional): The name of the column in the dataframe representing the time variable. Defaults to None.
        lookback_periods (int, optional): The number of periods to consider for lagged values. Defaults to 1.
        differencing_periods (int, optional): The number of periods to difference the time series. Defaults to 0.
        lag_periods (int, optional): The number of periods for the autoregressive component of the model. Defaults to 1.
        moving_average_periods (int, optional): The number of periods for the moving average component of the model. Defaults to 1.
        plot_time_series (bool, optional): Whether to plot the time series. Defaults to False.
        time_series_figure_size (tuple, optional): The size of the figure for the time series plot. Defaults to (8, 5).
        line_color (str, optional): The color of the line in the time series plot. Defaults to "#3269a8".
        line_alpha (float, optional): The transparency of the line in the time series plot. Defaults to 0.8.
        number_of_x_axis_ticks (int, optional): The number of ticks on the x-axis in the time series plot. Defaults to None.
        x_axis_tick_rotation (float, optional): The rotation angle of the x-axis tick labels in the time series plot. Defaults to None.
        title_for_plot (str, optional): The title of the time series plot. Defaults to "Time Series".
        subtitle_for_plot (str, optional): The subtitle of the time series plot. Defaults to "Shows the time series for the outcome variable.".
        caption_for_plot (str, optional): The caption of the time series plot. Defaults to None.
        data_source_for_plot (str, optional): The data source information for the time series plot. Defaults to None.
        x_indent (float, optional): The x-coordinate of the text indent in the time series plot. Defaults to -0.127.
        title_y_indent (float, optional): The y-coordinate of the title text indent in the time series plot. Defaults to 1.125.
        subtitle_y_indent (float, optional): The y-coordinate of the subtitle text indent in the time series plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-coordinate of the caption text indent in the time series plot. Defaults to -0.3.
        test_for_stationarity (bool, optional): Whether to test for stationarity of the time series. Defaults to True.
        show_acf_pacf_plots (bool, optional): Whether to plot the ACF and PACF plots. Defaults to False.
        show_model_results (bool, optional): Whether to show the model results. Defaults to False.
        plot_residuals (bool, optional): Whether to plot the residuals. Defaults to True.

    Returns:
        statsmodels.tsa.arima.model.ARIMAResults: The fitted ARIMA model.
    """
    
    # If time series plot is requested, ensure time column is provided
    if plot_time_series:
        if time_column_name == None:
            raise Exception('A time column must be provided to plot the time series.')
    
    # Conduct ADF test to determine if data is stationary
    if test_for_stationarity:
        adfuller_test = adfuller(dataframe[outcome_column_name])
        adfuller_pvalue = adfuller_test[1]
        if adfuller_pvalue < 0.05:
            print('The data is stationary.')
        else:
            warnings.warn("The outcome variable is not stationary. Consider differencing the data.")
    
    # Plot time series, if requested
    if plot_time_series:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=time_series_figure_size)
        
        # Use Seaborn to create a line plot
        sns.lineplot(
            data=dataframe,
            x=time_column_name,
            y=outcome_column_name,
            color=line_color,
            alpha=line_alpha,
            ax=ax
        )
        
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
        
        # Set the number of ticks on the x-axis
        if number_of_x_axis_ticks is not None:
            ax.xaxis.set_major_locator(plt.MaxNLocator(number_of_x_axis_ticks))
            
        # Rotate x-axis tick labels
        if x_axis_tick_rotation is not None:
            plt.xticks(rotation=x_axis_tick_rotation)

        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Word wrap the subtitle without splitting words
        if subtitle_for_plot != None:   
            subtitle_for_plot = textwrap.fill(subtitle_for_plot, 100, break_long_words=False)
            # Set the subtitle with Arial font, size 11, and color #666666
            ax.text(
                x=x_indent,
                y=subtitle_y_indent,
                s=subtitle_for_plot,
                fontsize=11,
                color="#666666",
                transform=ax.transAxes
            )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, 0.84)
        ax.yaxis.set_label_text(
            outcome_column_name,
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(0.9, -0.1)
        ax.xaxis.set_label_text(
            time_column_name,
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )

        # Show plot
        plt.show()
    
    # Plot ACF and PACF plots
    if show_acf_pacf_plots:
        plot_acf(dataframe[outcome_column_name], lags=lookback_periods)
        plt.show()
        try:
            plot_pacf(dataframe[outcome_column_name], lags=lookback_periods)
            plt.show()
        except:
            print('Cannot product PACF plot for the lookback periods.')
            pass
        
    # Create SARIMAX model
    arima_model = SARIMAX(
        dataframe[outcome_column_name],
        order=(lag_periods, differencing_periods, moving_average_periods),
    )
    
    # Fit the model
    arima_model = arima_model.fit()
    
    # Show model results, if requested
    if show_model_results:
        print(arima_model.summary())
        
    # Plot residuals, if requested
    if plot_residuals:
        arima_model.resid.plot(kind='kde')
        plt.show()
    
    # Return the model
    return arima_model




# Load packages
import numpy as np
import pandas as pd

# Declare function
def CreateStratifiedRandomSampleGroups(dataframe, 
                                       number_of_groups, 
                                       list_categorical_column_names, 
                                       random_seed=412):
    """
    Perform stratified random sampling on a DataFrame.

    Args:
        dataframe (pandas DataFrame): The input DataFrame.
        number_of_groups (int): The number of groups to create.
        list_categorical_column_names (list): A list of column names to use for stratification.
        random_seed (int, optional): The random seed for reproducibility (default is 412).

    Returns:
        pandas DataFrame: The DataFrame with an additional column 'group' indicating the group number.
    """
    # Ensure that the number of groups is a positive integer
    if number_of_groups <= 0:
        raise ValueError("Number of groups must be a positive integer (greater than 0)")
    
    # Ensure that the stratification variables are in the DataFrame
    if not all(col in dataframe.columns for col in list_categorical_column_names):
        raise ValueError("One or more stratification variables are not in the DataFrame")

    # Set the random seed for reproducibility
    np.random.seed(random_seed)

    # Group by the stratification variables
    grouped = dataframe.groupby(list_categorical_column_names, group_keys=True)

    # For each group, assign records to n groups
    def assign_to_group(sub_df):
        sub_df['group'] = (sub_df.sample(frac=1, random_state=random_seed).reset_index(drop=True).index % number_of_groups) + 1
        return sub_df

    # Apply the function to each group and combine the results
    stratified_df = grouped.apply(assign_to_group).reset_index(drop=True)
    
    # Return the updated DataFrame
    return stratified_df




# Load packages
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
from yellowbrick.cluster import KElbowVisualizer

# Declare function
def CreateKMeansClusters(dataframe,
                         list_of_value_columns_for_clustering=None,
                         number_of_clusters=None,
                         column_name_for_clusters='K-Means Cluster',
                         random_seed=412,
                         maximum_iterations=300,
                         scale_predictor_variables=True,
                         show_cluster_summary_plots=True,
                         color_palette='Set2',
                         # Text formatting arguments
                         caption_for_plot=None,
                         data_source_for_plot=None,
                         show_y_axis=False,
                         title_y_indent=1.1,
                         subtitle_y_indent=1.05,
                         caption_y_indent=-0.15,
                         # Plot formatting arguments
                         summary_plot_size=(6, 4)):
    """
    This function creates K-Means clusters on a dataset based on the variables specified.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        list_of_value_columns_for_clustering (list, optional): The list of variables to base the clusters on. Defaults to None, which will use all variables in the dataframe.
        number_of_clusters (int, optional): The number of clusters to create. Defaults to None, which will use the elbow method to determine the optimal number of clusters.
        column_name_for_clusters (str, optional): The name of the new column containing the clusters. Defaults to 'K-Means Cluster'.
        scale_predictor_variables (bool, optional): Whether to scale the predictor variables prior to analysis. Defaults to True.
        show_cluster_summary_plots (bool, optional): Whether to show cluster summary plots. Defaults to True.
        summary_plot_size (tuple, optional): The size of the summary plots. Defaults to (20, 20).
        random_seed (int, optional): The random seed to use for replication. Defaults to 412.
        maximum_iterations (int, optional): The maximum number of iterations to use for the K-Means algorithm. Defaults to 300.
    
    Returns:
        Pandas dataframe: An updated Pandas dataframe with the clusters joined to the original data.
    """
    # If no list of variables specified, use all numeric variables
    if list_of_value_columns_for_clustering == None:
        list_of_value_columns_for_clustering = list(dataframe.select_dtypes(include=np.number).columns)
    
    # Keep complete cases only
    dataframe_clusters = dataframe.dropna(subset=list_of_value_columns_for_clustering)

    # Scale the predictors, if requested
    if scale_predictor_variables:
        # Scale predictors
        dataframe_clusters[list_of_value_columns_for_clustering] = StandardScaler().fit_transform(dataframe_clusters[list_of_value_columns_for_clustering])
    
    # # Show peak-to-peak range of each variable
    # print("\nPeak-to-peak range of each value column/variable:")
    # print(np.ptp(dataframe_clusters[list_of_value_columns_for_clustering], axis=0))
    
    # If number of clusters not specified, use elbow to find "best" number
    if number_of_clusters == None:
        # Maximum number of clusters is 20 or the number of observations, whichever is smaller
        max_clusters = min(20, dataframe_clusters.shape[0])
        
        # Conduct elbow method
        model = KMeans()
        visualizer = KElbowVisualizer(
            model,
            k=(2, max_clusters),
            timings=True
        )
        visualizer.fit(dataframe)
        visualizer.show()
        number_of_clusters = visualizer.elbow_value_
        
    # Conduct k-means clusters
    model = KMeans(
        n_clusters=number_of_clusters, 
        random_state=random_seed,
        max_iter=maximum_iterations,
    )
    model = model.fit(dataframe_clusters[list_of_value_columns_for_clustering])

    # Join clusters to original dataset
    dataframe_clusters[column_name_for_clusters] = model.labels_

    # Convert cluster to categorical data type
    dataframe_clusters[column_name_for_clusters] = dataframe_clusters[column_name_for_clusters].apply(str)
    
    # Join clusters to original dataset
    dataframe_clusters = dataframe_clusters[[column_name_for_clusters]]
    dataframe = dataframe.merge(
        dataframe_clusters,
        how='left',
        left_index=True,
        right_index=True
    )
    
    if show_cluster_summary_plots:
        # Loop through each numeric variable
        for numeric_var in list_of_value_columns_for_clustering:
            # Create a temporary dataframe for the cluster
            dataframe_temp = dataframe[[column_name_for_clusters, numeric_var]]
            
            # Rename the variable to 'Value'
            dataframe_temp = dataframe_temp.rename(
                columns={numeric_var: 'Value'}
            )
            
            # Add the variable name to the dataframe
            dataframe_temp['Variable'] = numeric_var
            
            # Reorder the columns
            dataframe_temp = dataframe_temp[[column_name_for_clusters, 'Variable', 'Value']]
            
            # Create figure and axes
            fig, ax = plt.subplots(figsize=summary_plot_size)
                
            # Generate density plot using seaborn
            sns.kdeplot(
                data=dataframe_temp,
                x='Value',
                hue=column_name_for_clusters,
                fill=True,
                common_norm=False,
                alpha=.4,
                palette=color_palette,
                linewidth=1
            )
            
            # Remove top, left, and right spines. Set bottom spine to dark gray.
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['left'].set_visible(False)
            ax.spines['bottom'].set_color("#262626")
            
            # Remove the y-axis, and adjust the indent of the plot titles
            if show_y_axis == False:
                ax.axes.get_yaxis().set_visible(False)
                x_indent = 0.015
            else:
                x_indent = -0.005
            
            # Set the title with Arial font, size 14, and color #262626 at the top of the plot
            ax.text(
                x=x_indent,
                y=title_y_indent,
                s="Cluster Summary Plot",
                fontname="Arial",
                fontsize=14,
                color="#262626",
                transform=ax.transAxes
            )
            
            # Set the subtitle with Arial font, size 11, and color #666666
            ax.text(
                x=x_indent,
                y=subtitle_y_indent,
                s="Variable: " + numeric_var,
                fontname="Arial",
                fontsize=11,
                color="#666666",
                transform=ax.transAxes
            )
            
            # Set x-axis tick label font to Arial, size 9, and color #666666
            ax.tick_params(
                axis='x',
                which='major',
                labelsize=9,
                labelcolor="#666666",
                pad=2,
                bottom=True,
                labelbottom=True
            )
            plt.xticks(fontname='Arial')
            
            # Add a word-wrapped caption if one is provided
            if caption_for_plot != None or data_source_for_plot != None:
                # Create starting point for caption
                wrapped_caption = ""
                
                # Add the caption to the plot, if one is provided
                if caption_for_plot != None:
                    # Word wrap the caption without splitting words
                    wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                    
                # Add the data source to the caption, if one is provided
                if data_source_for_plot != None:
                    wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
                
                # Add the caption to the plot
                ax.text(
                    x=x_indent,
                    y=caption_y_indent,
                    s=wrapped_caption,
                    fontname="Arial",
                    fontsize=8,
                    color="#666666",
                    transform=ax.transAxes
                )
                
            # Show plot
            plt.show()
            
            # Clear plot
            plt.clf()

    # Return updated dataset
    return(dataframe)




# Load packages
import pandas as pd
from IPython.display import display, Markdown

# Declare function
def VerifyGranularity(dataframe,
                      list_of_key_columns,
                      set_key_as_index=True,
                      print_as_markdown=True):
    """
    Verifies the granularity of a given dataframe based on a list of key columns.
    This function creates a key column from the provided list of key columns and checks if the number of rows in the dataframe equals the number of distinct values in the key column. If the counts do not match, a warning message is printed. 
    The function can optionally set the key column as the dataframe's index and print the results as markdown.

    Args:
        dataframe (pd.DataFrame): The dataframe to verify.
        list_of_key_columns (list): The list of columns to use for creating the key column.
        set_key_as_index (bool, optional): If True, sets the key column as the dataframe's index. Defaults to True.
        print_as_markdown (bool, optional): If True, prints the results as markdown. Defaults to True.

    Returns:
        pd.DataFrame: The updated dataframe with the key column set as the index, if requested.
    """
    
    # Create key column from list of key columns
    first_col = list_of_key_columns[0]
    dataframe['Dataset Key'] = dataframe[first_col].astype(str)
    
    # If more than 1 column listed, concatenate columns together
    if len(list_of_key_columns) > 1:
        for key_col in list_of_key_columns[1:]:
            dataframe['Dataset Key'] = dataframe['Dataset Key'].astype(str) + " -- " + dataframe[key_col].astype(str)
    
    # Get row count and distinct key count
    row_count = len(dataframe.index)
    distinct_key_count = len(pd.unique(dataframe['Dataset Key']))
    
    # Create row count string, format the count with thousands separator, and print it
    row_count_string = "Number of rows in dataframe: " + "{:,}".format(row_count)
    # Create distinct key count string, format the count with thousands separator, and print it
    distinct_key_count_string = "Distinct count of Dataset Keys: " + "{:,}".format(distinct_key_count)
    
    # Show granularity results as markdown if requested
    if print_as_markdown:
        # Show the results using check mark emoji if row count equals distinct key count
        display(Markdown(row_count_string))
        display(Markdown(distinct_key_count_string))
        if row_count != distinct_key_count:
            display(Markdown("⚠️ WARNING: Unique combination of columns you listed does not equal number of rows in dataset. Try a new combination of columns or see if dataset has duplicates."))
        else:
            display(Markdown("✅ SUCCESS: Dataset granularity is verified using the columns you listed."))
    else:
        print(row_count_string)
        print(distinct_key_count_string)
        if row_count != distinct_key_count:
            print("WARNING: Unique combination of columns you listed does not equal number of rows in dataset. Try a new combination of columns or see if dataset has duplicates.")
        else:
            print("SUCCESS: Dataset granularity is verified using the columns you listed.")
        
    # If Key column not being kept, drop it. Otherwise, move Key column to front of DataFrame
    if set_key_as_index != True:
        dataframe = dataframe.drop(columns=['Dataset Key'])
    else:
        dataframe.index = dataframe['Dataset Key']
        dataframe = dataframe.drop(columns=['Dataset Key'])
        return(dataframe)




# Load packages
import pandas as pd

# Declare function
def CreateRandomSampleGroups(dataframe,
                             number_of_groups=2, 
                             random_seed=412):
    """
    Shuffle the DataFrame rows, assign each record to one of n groups, then restore the original order and return the updated DataFrame.

    Args:
        dataframe (pandas.DataFrame): The input DataFrame.
        number_of_groups (int, optional): The number of groups to create. Defaults to 2.
        random_seed (int, optional): The random seed for reproducibility. Defaults to None.

    Returns:
        pandas.DataFrame: The updated DataFrame with an additional column 'group' indicating the group number.
    """
    # Ensure that the number of groups is a positive integer
    if number_of_groups <= 0:
        raise ValueError("Number of groups must be a positive integer (greater than 0)")

    # Shuffle the DataFrame
    shuffled_df = dataframe.sample(frac=1, random_state=random_seed).reset_index(drop=True)

    # Assign groups to the shuffled DataFrame
    group_numbers = [i % number_of_groups + 1 for i in range(len(shuffled_df))]
    shuffled_df['group'] = group_numbers
    
    # Return the updated DataFrame
    return shuffled_df




# Load packages
import pandas as pd
import numpy as np

# Declare function
def CreateSIPDataframe(name_of_items,
                       list_of_items,
                       number_of_trials=10000):
    """
    Creates a pandas DataFrame of trials of simulations (i.e., stochastic information packet, or SIP) for each item specified in a list.

    Args:
        name_of_items (str): The name of the item being simulated.
        list_of_items (list): A list of items to simulate.
        number_of_trials (int, optional): The number of trials to simulate. Defaults to 10000.

    Returns:
        pandas.DataFrame: A DataFrame containing the simulation results.
    """
    
    # Create dataframe of trials
    df_sips = pd.DataFrame(columns=[
        'Trial',
        name_of_items
    ])
    
    # Iterate through list of items
    for item in list(set(list_of_items)):
        # Create trial column with item name
        df_temp = pd.DataFrame({
            'Trial': np.linspace(1, number_of_trials, number_of_trials, dtype='int'),
            name_of_items: item
        })
        # Append to SIP dataframe
        df_sips = pd.concat([df_sips, df_temp])
        
    # Return SIP DataFrame
    return(df_sips)




# Load packages
from statsmodels.stats.oneway import anova_oneway
from statsmodels.stats.multicomp import pairwise_tukeyhsd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import ttest_1samp
import seaborn as sns
import textwrap

# Declare function
def OneWayANOVA(dataframe,
                outcome_column,
                grouping_column,
                signficance_level=0.05,
                homogeneity_of_variance=True,
                use_welch_correction=False,
                plot_sample_distributions=True,
                color_palette='Set2',
                title_for_plot="One-Way ANOVA",
                subtitle_for_plot="Shows the distribution of the sample means for each group.",
                caption_for_plot=None,
                data_source_for_plot=None,
                show_y_axis=False,
                title_y_indent=1.1,
                subtitle_y_indent=1.05,
                caption_y_indent=-0.15,
                figure_size=(8, 6)):
    """
    Conducts a one-way ANOVA on a given dataframe, with options for post-hoc testing and visualization.

    Args:
        dataframe: A pandas dataframe containing the data to be analyzed.
        outcome_column: A string representing the name of the column containing the outcome variable.
        grouping_column: A string representing the name of the column containing the grouping variable.
        signficance_level: A float representing the desired significance level (default: 0.05).
        homogeneity_of_variance: A boolean indicating whether to assume homogeneity of variance (default: True).
        use_welch_correction: A boolean indicating whether to use Welch's correction for unequal variances (default: False).
        plot_sample_distributions: A boolean indicating whether to plot the distribution of sample means for each group (default: True).
        color_palette: A string representing the name of the color palette to use for the plot (default: 'Set2').
        title_for_plot: A string representing the title of the plot (default: 'One-Way ANOVA').
        subtitle_for_plot: A string representing the subtitle of the plot (default: 'Shows the distribution of the sample means for each group.').
        caption_for_plot: A string representing the caption of the plot (default: None).
        data_source_for_plot: A string representing the data source of the plot (default: None).
        show_y_axis: A boolean indicating whether to show the y-axis on the plot (default: False).
        title_y_indent: A float representing the y-axis position of the title (default: 1.1).
        subtitle_y_indent: A float representing the y-axis position of the subtitle (default: 1.05).
        caption_y_indent: A float representing the y-axis position of the caption (default: -0.15).
        figure_size: A tuple representing the size of the plot (default: (8, 6)).

    Returns:
        A string representing the results of Tukey's test.
    """
    
    # Show pivot table summary
    table_groups = pd.pivot_table(dataframe, 
        values=[outcome_column], 
        index=[grouping_column],
        aggfunc={outcome_column: [len, min, max, np.mean, np.std, np.var]}
    )
    print(table_groups)
    print("\n")

    # Set variance argument
    if homogeneity_of_variance:
        variance_arg = "equal"
    else:
        variance_arg = "unequal"

    # Conduct one-way ANOVA
    test_results = anova_oneway(
        data=dataframe[outcome_column], 
        groups=dataframe[grouping_column], 
        use_var=variance_arg, 
        welch_correction=use_welch_correction
    )

    # Print interpretation of results
    if test_results.pvalue <= signficance_level:
        print("There is a statistically detectable difference between the sample means of at least 2 groups.")
    else:
        print("Cannot reject the null hypothesis.")
    print("p-value:", str(round(test_results.pvalue, 3)))
    print("degrees of freedom between groups:", str(int(test_results.df_num)))
    print("degrees of freedom across all groups:", str(int(test_results.df_denom)))
    print("f-statistic:", str(round(test_results.statistic, 3)))
    print("\n")
    
    # Perform Tukey's test
    tukey = pairwise_tukeyhsd(
        endog=dataframe[outcome_column],
        groups=dataframe[grouping_column],
        alpha=signficance_level
    )
    results = tukey.summary()
    
    # If plot_distribution_of_differences is True, plot distribution of differences
    if plot_sample_distributions:
        # Create placeholder for results
        data_distribution = pd.DataFrame()
        
        # Get unqiue groups
        list_of_groups = dataframe[grouping_column].unique()
        
        # Iterate through groups
        for group in list_of_groups:
            # Get the group's data
            group_data = dataframe[dataframe[grouping_column] == group][outcome_column]
            
            # Get the group's mean
            group_mean = group_data.mean()
            
            # Get the group's standard error
            group_se = group_data.sem()
            
            # Generate a normal distribution
            group_distribution = np.random.normal(group_mean, group_se, 10000)
            
            # Convert to dataframe
            df_group_distribution = pd.DataFrame(group_distribution,
                                                 columns=[outcome_column])
            
            # Add group name
            df_group_distribution[grouping_column] = group
            
            # Row bind to data_distribution
            data_distribution = pd.concat([data_distribution, df_group_distribution], axis=0).reset_index(drop=True)
            
        # Plot distribution of means
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
            
        # Generate density plot using seaborn
        sns.kdeplot(
            data=data_distribution,
            x=outcome_column,
            hue=grouping_column,
            fill=True,
            common_norm=False,
            alpha=.4,
            palette=color_palette,
            linewidth=1
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()

    # Return results
    return results




from .CreateARIMAModel import CreateARIMAModel
from .CreateBoostedTreeModel import CreateBoostedTreeModel
from .CreateDecisionTreeModel import CreateDecisionTreeModel
from .CreateLinearRegressionModel import CreateLinearRegressionModel
from .CreateLogisticRegressionModel import CreateLogisticRegressionModel
from .CreateNeuralNetwork_SingleOutcome import CreateNeuralNetwork_SingleOutcome


# Load packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from textwrap import wrap

# Declare function
def CreateDataOverview(dataframe,
                       plot_missingness=False):
    """
    This function creates an overview of the data in a dataframe, showing the data type, missing count, missing percentage, and summary statistics for each variable.
    Tip: This function is useful for creating a data dictionary.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        plot_missingness (bool, optional): Generates a plot to show missingness in each variable. Defaults to False.

    Returns:
        Pandas dataframe: A Pandas dataframe containing an overview of the data in the dataframe.
    """
    
    # Get data types in each column
    data_overview = pd.DataFrame(dataframe.dtypes, columns=['DataType'])
    data_overview = data_overview.reset_index()
    data_overview = data_overview.rename(columns={
        'index': 'Variable',
        'DataType': 'Data Type'
    })
    
    # Count missing values in each column
    data_missing = dataframe.isnull().sum().reset_index()
    data_missing = data_missing.reset_index()
    data_missing = data_missing.rename(columns={
        'index': 'Variable',
        0: 'Missing Count'
    })
    data_missing = data_missing[['Variable', 'Missing Count']]
    
    # Join missing count to the overview
    data_overview = data_overview.merge(
        data_missing,
        how='left',
        on='Variable'
    )
    del(data_missing)
    
    # Calculate missing percentage
    data_overview['Missing Percentage'] = data_overview['Missing Count'] / len(dataframe)
    
    # Show range and frequency in each column
    data_summary = dataframe.describe(include='all', datetime_is_numeric=True).T
    data_summary = data_summary.reset_index()
    data_summary.columns.values[0] = 'Variable'
    try:
        data_summary = data_summary[['Variable', 'count', 'unique', 'top', 'freq', 'min', 'max']]
        data_summary = data_summary.rename(columns={
            'count': 'Non Missing Count',
            'unique': 'Unique Value Count',
            'top': 'Top Value',
            'freq': 'Frequency of Top Value',
            'min': 'Minimum',
            'max': 'Maximum'
        })
    except KeyError:
        try:
            data_summary = data_summary[['Variable', 'count', 'min', 'max']]
            data_summary = data_summary.rename(columns={
                'count': 'Non Missing Count',
                'min': 'Minimum',
                'max': 'Maximum'
            })
        except KeyError:
            data_summary = data_summary[['Variable', 'count', 'unique', 'top', 'freq']]
            data_summary = data_summary.rename(columns={
                'count': 'Non Missing Count',
                'unique': 'Unique Value Count',
                'top': 'Top Value',
                'freq': 'Frequency of Top Value'
            })
    
    # Join the two dataframes
    data_overview = data_overview.merge(
        data_summary, 
        how='left',
        on='Variable')
    del(data_summary)
    
    # Calculate the range of each column
    try:
        data_overview['Range'] = data_overview['Maximum'] - data_overview['Minimum']
    except KeyError:
        pass
    
    # Generate missingness plot, if requested
    if plot_missingness:
        # Sort the data by missing percentage
        data_overview_sorted = data_overview.sort_values(
            by=['Missing Percentage', 'Variable'],
            ascending=[False, True]
        )
        # Create the plot
        fig, ax = plt.subplots(figsize=(8, len(data_overview_sorted.index)))
        sns.barplot(
            data=data_overview_sorted, 
            y='Variable', 
            x='Missing Percentage',
            palette='Blues_d'
        )
        # Add data labels
        lbls = [f"{x:.0%}" for x in data_overview_sorted['Missing Percentage']]
        ax.bar_label(container=ax.containers[0],
                     labels=lbls,
                     padding=5)
        # Set y-axis limits and hide ticks
        plt.xlim(0, 1)
        plt.xticks([])
        # Hide the spines
        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['bottom'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        # Format the x-axis labels and wrap them
        plt.gca().yaxis.set_tick_params(labelsize=8)
        y_labels = plt.gca().get_yticklabels()
        wrapped_labels = []
        for label in y_labels:
            wrap_label = '\n'.join(wrap(label.get_text(), 30))
            wrapped_labels.append(wrap_label)
        plt.gca().set_yticklabels(wrapped_labels)
        # Show the plot
        plt.show()
    
    # Return the overview
    return(data_overview)




# Load packages
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import statsmodels.api as sm
import seaborn as sns

# Declare function
def ConductLogisticRegressionAnalysis(dataframe,
                                      outcome_variable,
                                      list_of_predictors,
                                      scale_predictors=False,
                                      show_diagnostic_plots_for_each_predictor=True,
                                      show_help=True):
    """
    Conducts a logistic regression analysis on the specified dataframe, using the specified outcome variable and list of predictors.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be analyzed.
        outcome_variable (str): The name of the column in the dataframe containing the outcome variable.
        list_of_predictors (list of str): A list of the names of the columns in the dataframe containing the predictor variables.
        scale_predictors (bool, optional): Whether or not to scale the predictor variables. Default is False.
        show_diagnostic_plots_for_each_predictor (bool, optional): Whether or not to show diagnostic plots for each predictor variable. Default is True.
        show_help (bool, optional): Whether or not to show help text. Default is True.

    Returns:
        dict: A dictionary containing the fitted model and model summary.
    """
    
    # Select columns specified
    dataframe = dataframe[list_of_predictors + [outcome_variable]].copy()
    
    # Remove NAN and inf values
    dataframe.dropna(inplace=True)
    dataframe = dataframe[np.isfinite(dataframe).all(1)]

    # Add constant
    dataframe = sm.add_constant(dataframe)
    
    # Scale the predictors, if requested
    if scale_predictors:
        # Show the mean and standard deviation of each predictor
        print("\nMean of each predictor:")
        print(dataframe[list_of_predictors].mean())
        print("\nStandard deviation of each predictor:")
        print(dataframe[list_of_predictors].std())
        
        # Scale predictors
        dataframe[list_of_predictors] = StandardScaler().fit_transform(dataframe[list_of_predictors])
    
    # Get number of unique outcomes
    number_of_unique_outcomes = len(dataframe[outcome_variable].unique())
    
    # If there are more than two outcomes, conduct multinomial regression
    if number_of_unique_outcomes > 2:
        # Create multinomial regression model
        model = sm.MNLogit(dataframe[outcome_variable], dataframe[['const'] + list_of_predictors])
        try:
            model_res = model.fit()
        except:
            model = sm.MNLogit(dataframe[outcome_variable], dataframe[list_of_predictors])
            model_res = model.fit()
    else:
        # Create binomial regression model
        model = sm.Logit(dataframe[outcome_variable], dataframe[['const'] + list_of_predictors])
        try:
            model_res = model.fit()
        except:
            model = sm.Logit(dataframe[outcome_variable], dataframe[list_of_predictors])
            model_res = model.fit()
    model_summary = model_res.summary()
    
   # If requested, show diagnostic plots
    if show_diagnostic_plots_for_each_predictor:
        for variable in list_of_predictors:
            fig = plt.figure(figsize=(12, 8))
            sns.regplot(
                x=variable, 
                y=outcome_variable, 
                data=dataframe,
                logistic=True, 
                y_jitter=.03
            )
            # Show plots
            plt.show()
    
    # If requested, show help text
    if show_help:
        print(
            "Quick guide on accessing output of ConductLogisticRegressionAnalysis function:",
            "\nThe ouput of the ConductLogisticRegressionAnalysis function is a dictionary containing the regression results, a test dataset of predictors, and a test dataset of outcomes.",
            "\n\t--To access the logistic regression model, use the 'Fitted Model' key."
            "\n\t--To view the model's statistical summary, use the 'Model Summary' key."
        )
    
    # Create dictionary of objects to return
    dict_return = {
        "Fitted Model": model_res,
        "Model Summary": model_summary
    }
    return dict_return

